
#include "shared.h"
#include "LJZmenu.h"

#include "filters.h"

#include "PenInputMgr.h"


// Assertion Support
static void
assert_check(int error, const char* message)
{
    if (error) {
        ErrDisplayFileLineMsg("PalmOS", (UInt16) error, message);
    }
}

#define xassert(error) {assert_check(error, #error);}

static void showError(const char* aWhat, Err aError) {
    ErrDisplayFileLineMsg("oink", (UInt16) aError, aWhat);
}

extern UInt32 mStopEmulation;
extern UInt32 device_type;

char str_month[12][4]=
{"Jan","Feb","Mar","Apr","May","Jun",
"Jul","Aug","Sep","Oct","Nov","Dec"};

char RomPath[260];
char shortrom_filename[256];
char genZ_curVolume;

int genZ_loadrom();
void genZ_end();


void genZ_changefskip(void);
void genZ_changerendering(void);
void genZ_updatevpad(unsigned int val);
void genZ_MainLoop(void);
void genZ_pause(int pause);
int genZ_ShowMenu(void);
void genZ_updateScreen(void);
void genZ_asmfastrender256(uint8 *screen,uint8 *source,uint32 h,uint32 ofs);
void genZ_asmfastrender320(uint8 *screen,uint8 *source,uint32 h,uint32 ofs);


uint32 genZ_enabledSound,genZ_sndfreq;
uint32 genZ_renderingmode,genZ_smoothmode,genZ_flipping;
uint32 genZ_padState;
int32 genZ_frameskipped,genZ_frameskip,genZ_fastforward,genZ_specialfskip;
int32 genZ_frameBeginTime;
int32 genZ_speedpct;
int32 genZ_autofireA,genZ_autofireB,genZ_autofireC;
int32 genZ_framecount,genZ_drawframe;

int genZ_cpu_cyclone;

int32 genZ_frameTime;

UInt16 *genZ_zodvram;

int state_load(int num,int onlytest);
int state_save(int num);

uint8 dummy_scale2xsai1[512*3*2];
uint8 VideoBuffer[512*512*2];
uint8 dummy_scale2xsai2[512*3*2];


static unsigned char *RomData;
static int RomLen;

#include "SoundMgr.h"
static SndStreamRef snd_stream=NULL;
Int32 snd_stream_vol;

typedef struct {uint32 __r9;uint32 __r10;} reg_sys_t;
static reg_sys_t reg_sys;
uint32 genZ_sndbuffsize;


static Err genZ_sndCallBack (void *userData,SndStreamRef stream, void *bufferZ,UInt32 frameCountZ)
{
	__asm 
	{
		stmfd  r13!,{r9,r10}
		ldr	   r9,[r0]
		ldr	   r10,[r0,#4]
	}		
	//fm_update_timers();
	audio_update((int16*)bufferZ);
	__asm 
	{
		ldmfd  r13!,{r9,r10}
	}
	return errNone;
}

int update_input(void);

extern UInt32 ljp_isLowRes,ljp_confirmOW,ljp_confirmSL;

void emu_resettimer(void) {
	genZ_frameBeginTime=GetTickCount()+17*100/genZ_speedpct;
}

void genZ_togglesound(void) { //******** TODO **************
	char tmp[64];
	int i;
	
	genZ_pause(1);
	
	/*genZ_enabledSound++;
	if (genZ_enabledSound>2) genZ_enabledSound=0;
	
	if (genZ_enabledSound==0) strcpy(tmp,"Sound : OFF");
	else if (genZ_enabledSound==1) strcpy(tmp,"Sound : SIM");
	else if (genZ_enabledSound==2) strcpy(tmp,"Sound : ON");*/
	
	

	menu_message(tmp);
	
	LJP_WAIT(300);
	emu_resettimer();
	
	genZ_pause(0);
}

char genZ_reset(void) {
	genZ_pause(1);
	
  	if (menu_confirm("Reset Genesis ?")) {gen_reset(); /*return 1;*/}
  	
  	emu_resettimer();
  	
  	genZ_pause(0);
  	
  	return 0;
}

void genZ_quicksave(void) {
	genZ_pause(1);
	
	if (CONFIRM_SAVE) state_save(0);
	emu_resettimer();
	ClearScreen();
	
	genZ_pause(0);
}

void genZ_quickload(void) {
	genZ_pause(1);
	
	if (!state_load(0,1))
	{
		if (CONFIRM_LOAD) state_load(0,0);
		emu_resettimer();
		ClearScreen();
	}
	
	genZ_pause(0);
}

void load_sram(void)
{
	FILE *f;	
	char savename[255],result[256];
	uint8 i=0,j=255,k=255,l=255;
	char chaine[256];  
	i=0;
	j=255;
	k=-1;
	while (RomPath[i]!=0) 
	{
		if (RomPath[i]=='.') j=i;
		if (RomPath[i]=='/') k=i;
		if (RomPath[i]=='\\') k=i;
		i++;
	  }
	if (j==255) j=i;  //pas d'extension	  
	memcpy(result,RomPath+k+1,j+1-k-1);    
	result[j+1-k-1]=0;
	strcat(result,"srm");
	sprintf(savename,"%s%s",GENZ_SAVESDIR,result);
	savename[0]=genZ_curVolume+'0';
	
	f=fopen(savename,"rb");
	if (!f) return;
	fread(saveram,save_len,1,f);
	fclose(f);
}
void save_sram(void)
{
	FILE *f;	
	char savename[255],result[256];
	uint8 i=0,j=255,k=255,l=255;
	char chaine[256];  
	i=0;
	j=255;
	k=-1;
	while (RomPath[i]!=0) 
	{
		if (RomPath[i]=='.') j=i;
		if (RomPath[i]=='/') k=i;
		if (RomPath[i]=='\\') k=i;
		i++;
	  }
	if (j==255) j=i;  //pas d'extension	  
	memcpy(result,RomPath+k+1,j+1-k-1);    
	result[j+1-k-1]=0;
	strcat(result,"srm");
	sprintf(savename,"%s%s",GENZ_SAVESDIR,result);
	savename[0]=genZ_curVolume+'0';
	f=fopen(savename,"w+b");
	if (!f) return;
	fwrite(saveram,save_len,1,f);
	fclose(f);
}

int update_input(void)
{
    int running = 1;
    
  int pad=0;
  
  if (genZ_padState&KEY_QUICKLOAD)
	{
		if (!state_load(0,1))
		{
			if (CONFIRM_LOAD) state_load(0,0);
			genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
		}
	}
	if (genZ_padState&KEY_QUICKSAVE)
	{
		if (CONFIRM_SAVE) state_save(0);
			genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
	}
	if (genZ_padState&KEY_TURBO)
	{		
		genZ_fastforward=1;		
	}
	else 
	if (genZ_fastforward)
	{
		genZ_fastforward=0;
		genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
	}
    
  if(genZ_padState&KEY_UP)     pad|=INPUT_UP;
  if(genZ_padState&KEY_DOWN)   pad|=INPUT_DOWN;
  if(genZ_padState&KEY_LEFT)   pad|=INPUT_LEFT;
  if(genZ_padState&KEY_RIGHT)  pad|=INPUT_RIGHT;
  if(genZ_padState&KEY_A) 
  {
	 switch (genZ_autofireA)
	 {
  		case 0:pad|=INPUT_A; break;
  		case 1:
  	  		if ((genZ_framecount&31)>15) pad|=INPUT_A;
  	  		break;
  		case 2:
  	  		if ((genZ_framecount&15)>7) pad|=INPUT_A;
  	  		break;
  		case 3:
  	  		if ((genZ_framecount&7)>3) pad|=INPUT_A;
  	  		break;  		   
  	 }
  }  		    
  if(genZ_padState&KEY_B)
  {
 	 switch (genZ_autofireB)
	 {
  		case 0:pad|=INPUT_B; break;
  		case 1:
  	  		if ((genZ_framecount&31)>15) pad|=INPUT_B;
  	  		break;
  		case 2:
  	  		if ((genZ_framecount&15)>7) pad|=INPUT_B;
  	  		break;
  		case 3:
  	  		if ((genZ_framecount&7)>3) pad|=INPUT_B;
  	  		break;  		   
  	 }
  }
  if(genZ_padState&KEY_C)
  {
 	 switch (genZ_autofireC)
	 {
  		case 0:pad|=INPUT_C; break;
  		case 1:
  	  		if ((genZ_framecount&31)>15) pad|=INPUT_C;
  	  		break;
  		case 2:
  	  		if ((genZ_framecount&15)>7) pad|=INPUT_C;
  	  		break;
  		case 3:
  	  		if ((genZ_framecount&7)>3) pad|=INPUT_C;
  	  		break;  		   
  	 }
  }
  if(genZ_padState&KEY_START)  pad|=INPUT_START;

  input.pad[0]=pad;
/*    if(keystate[SDLK_z])      input.pad[0] |= INPUT_X;
    if(keystate[SDLK_x])      input.pad[0] |= INPUT_Y;
    if(keystate[SDLK_c])      input.pad[0] |= INPUT_Z;
    if(keystate[SDLK_v])      input.pad[0] |= INPUT_MODE;*/
 
   //system_reset();

    //if(keystate[SDLK_ESCAPE]) running = 0;
    return (running);
}



static int Do_Nothing(void)
{
	return 0;
}


/*******************************************************/


int init_video(void)
{
	//Init vid
		
	signed long val1,val2,val;
	FtrGet(CREATORID,FTR_OPT_SCREENMODE_ID,(unsigned long*)&val1);
	FtrGet(CREATORID,FTR_OPT_SMOOTHING_ID,(unsigned long*)&val2);
	genZ_renderingmode=val1;
	genZ_smoothmode=val2;
	
	if (genZ_smoothmode==2){
		if (InitStretchTables()){
			menu_info("Err:InitStretchTables");
			genZ_smoothmode=0;
		}
	}

	

	genZ_fastforward=0;
	
	genZ_frameskipped=0;
	genZ_specialfskip=0;
	
	 FtrGet(CREATORID,FTR_OPT_FSKIP_ID,(unsigned long*)&val); 
	genZ_frameskip=val;
	if (genZ_frameskip==SKEEZIX_FSKIP) genZ_specialfskip=1;
	

    filters_init();

	return 0;
}


int genZ_loadrom()
{
	ljz_std_showstat();
	int val;
				
	FtrGet(CREATORID,FTR_OPT_SOUND_ID,(unsigned long*)&val);  
	genZ_enabledSound=val;
	FtrGet(CREATORID,FTR_OPT_CPUCORE_ID,(unsigned long*)&val);  
	genZ_cpu_cyclone=val;	
	
	FtrGet(CREATORID,FTR_OPT_SOUNDFREQ_ID,(unsigned long*)&val);  
	if (val==1) genZ_sndfreq=22050;
	else if (val==2) genZ_sndfreq=11025;
	else genZ_sndfreq=44100;
	
	FtrGet(CREATORID,FTR_OPT_SOUNDVOLUME_ID,(unsigned long*)&val); 
	snd_stream_vol=val;
	if (snd_stream_vol<0) snd_stream_vol=0;
	if (snd_stream_vol>128) snd_stream_vol=128;
	snd_stream_vol*=16;  
	
	emul_z80=(genZ_enabledSound&4);
	
	genZ_speedpct=100;
	genZ_autofireA=0;
	genZ_autofireB=0;
	genZ_autofireC=0;
	
	genZ_framecount=0;
	genZ_drawframe=1;
	



	init_video();
	
    char *ftrptr;
    strcpy(RomPath,GENZ_ROMSDIR);    
    
    {//handle categories
    	char cat[32];
	  	FtrGet(CREATORID,FTR_OPT_CATEGORY_ID ,(unsigned long*)&ftrptr);
	  	strcpy(cat,ftrptr);
  		if ( StrCaselessCompare(cat,"uNaRcHiVeD") != 0) {
  			strcat(RomPath,cat);
	  		strcat(RomPath,"/");
	  	}
	  	//FtrPtrFree(CREATORID, FTR_OPT_CATEGORY_ID);
  	}
    
    FtrGet(CREATORID,FTR_ROM_FILENAME_ID ,(unsigned long*)&ftrptr);
    
    genZ_curVolume=ftrptr[0];
    RomPath[0]=genZ_curVolume+'0';
    
    ftrptr+=3;
    strcpy(shortrom_filename,ftrptr);  
    strcat(RomPath,ftrptr);
//    FtrPtrFree(CREATORID, FTR_ROM_FILENAME_ID);	
	
	// Init emulation:		
    error_init();

	/* Load game */
//  char strload[128];
//    sprintf(strload,"Loading : %s",shortrom_filename);
    menu_message(shortrom_filename/*strload*/);

//		unsigned long freeM,maxM;	
//	MemHeapFreeBytes(0,&freeM,&maxM);
//	StrPrintF(strload,"mem : %d %d",freeM,maxM);
//	menu_message(strload);
//	menu_waitPen(NULL,NULL,1);


    if(!load_rom(RomPath))
	{
		char caption[256];
		sprintf(caption, "Error loading file `%s'.",RomPath);
//		MessageBox(NULL, caption, "Error", 0);
		menu_message(caption);
		menu_waitPen(NULL,NULL,1);
//		exit(1);
		return 1;
	}

	if (saveram)
	{
		//char str[256];		
		//sprintf(str,"found sram:sz=%d,st=%08X",save_len,save_start);
		//menu_message(str);
		//menu_waitPen(NULL,NULL,1);
		load_sram();
	}

    memset(&bitmap, 0, sizeof(t_bitmap));
    bitmap.width  = 512;
    bitmap.height = 512;
    bitmap.depth  = 16;
    bitmap.granularity = 2;
    bitmap.pitch = (bitmap.width * bitmap.granularity);
    bitmap.data   = (unsigned char *)VideoBuffer;
    bitmap.viewport.w = 256;
    bitmap.viewport.h = 224;
    bitmap.viewport.x = 0x20;
    bitmap.viewport.y = 0x00;
    bitmap.remap = 1;

    system_init();

    genZ_sndbuffsize=1024;    
    if (genZ_enabledSound&2) 
    {
    	audio_init(genZ_sndfreq,genZ_sndbuffsize);
    	audio_fastmode(genZ_enabledSound&8);
    }

    __asm 
    {
  		ldr  r0, = reg_sys
	  	add	 r0,r0,r10
  		str	 r9,[r0]
	  	str	 r10,[r0,#4]
    } 
    

    system_reset();	

    
    genZ_frameTime=(frame_end==0xE0?17:20); //hardcoded to NTSC
            
    genZ_frameBeginTime=GetTickCount();
    

	genZ_pause(0);

    return 0;  
}
void genZ_end()
{
      int val;
	 genZ_pause(1);


/*emu*/	  
	  if (saveram) 
	  {
	  	save_sram();
	  	ljz_free(saveram);
	  }

 	  system_shutdown();
      error_shutdown();
      
    FtrPtrFree(CREATORID, FTR_ROM_ID);                  
            

	FtrSet(CREATORID, FTR_OPT_SMOOTHING_ID,genZ_smoothmode);	

	FtrSet(CREATORID, FTR_OPT_SOUND_ID,genZ_enabledSound);
	switch (genZ_sndfreq)
	{
		case 11025:val=2;break;
		case 22050:val=1;break;
		default:val=0;break;
	}
	FtrSet(CREATORID, FTR_OPT_SOUNDFREQ_ID,val);
	
	FtrSet(CREATORID,FTR_OPT_SOUNDVOLUME_ID,snd_stream_vol/16); 
	
	FtrSet(CREATORID, FTR_OPT_SCREENMODE_ID,genZ_renderingmode);
	
	FtrSet(CREATORID, FTR_OPT_FSKIP_ID,genZ_frameskip);
	ljz_std_showstat();	
}


void genZ_updateScreen(void)
{	 
	if (genZ_drawframe)
	{
		int width = (reg[12] & 1) ? 320 : 256;
	    int height = (reg[1] & 8) ? 240 : 224;
	    int y;
	    uint8 *dst,*src;
	    src = (uint8*)VideoBuffer+(bitmap.viewport.x<<1);
	    if (width==256) {       
		  if (!ljp_isLowRes) {	
			switch (genZ_renderingmode)	{
				case 0:
				case 2:	{
					dst = (uint8*)(g_pnobridge->vram)+((((display_width-width)>>1)+((display_height-height-(genZ_renderingmode<2?BELTBAR_SIZE:0))>>1)*display_pitch)<<1);					
					ljp_asmfastrender256(dst,src,height,(display_pitch-256)<<1);
					break;
				}
				case 1:
				case 3:	{
					
					if (device_type&BF_ZODIAC){				
						if (genZ_smoothmode&_ZOD_SMOOTHING_) {
							dst = ((uint8*) genZ_zodvram);
							ljp_asmfastrender256(dst,src,height,(480-256)<<1);
							stubzodiacblit(3,width,height,4);
							break;
						}
						else if (!genZ_smoothmode) {
							dst = ((uint8*) genZ_zodvram);
							ljp_asmfastrender256(dst,src,height,(480-256)<<1);
							stubzodiacblit(3,width,height,0);
							break;
						}
					}
					dst = (uint8*)(g_pnobridge->vram)+((((display_width-width*5/4)>>1)+((display_height-height*5/4-(genZ_renderingmode<2?BELTBAR_SIZE:0))>>1)*display_pitch)<<1);
					if (genZ_smoothmode==0) ljp_asmfastrenderX15256(dst,src,height,(display_pitch-320)<<1);
					else if (genZ_smoothmode==1) ljp_asmfastrenderX15sm256(dst,src,height,(display_pitch-320)<<1);
					else if (genZ_smoothmode==2) StretchNesterTrue(src,dst,512,display_pitch,height);
					else if (genZ_smoothmode==3) fgbZ_filterrender(dst,src,(UInt32)((display_pitch)*2),512*2,SCALE_2XSAI,256,height,320,height*5/4);
					break;
				}
				case 4:					
					if (device_type&BF_ZODIAC){						
						if (genZ_smoothmode&_ZOD_SMOOTHING_) {
							dst = ((uint8*) genZ_zodvram);						
							ljp_asmfastrender256(dst,src,height,(480-256)<<1);
							stubzodiacblit(3,256,height,5);
							break;
						} else if (!genZ_smoothmode) {
							dst = ((uint8*) genZ_zodvram);						
							ljp_asmfastrender256(dst,src,height,(480-256)<<1);
							stubzodiacblit(3,256,height,1);
							break;
						}
					}
					if (device_type&BF_HIRESPLUS){
						dst = (uint8*)(g_pnobridge->vram);//+((((display_width-width)>>1)+((display_height-height)>>1)*display_pitch)<<1);
						if (genZ_smoothmode) fgbZ_filterrender(dst,src,(UInt32)((display_pitch)*2),512*2,SCALE_2XSAI,256,height,480,320/*-BELTBAR_SIZE*/);
					}
					break;
			}
		  } else {
		  	dst = ((UInt16*) g_pnobridge -> vram) + ( ((g_pnobridge -> display_height-140-BELTBAR_SIZE)>>1) * display_pitch ) + ((display_width-160)>>1);
			fgbZ_filterrender((uint8*)dst,(uint8*)src,display_pitch*2,512*2,SCALE_2XSAI,256,height,160,140);
		  }
		}
		else {    //width==320
		  if (!ljp_isLowRes) {	
			switch (genZ_renderingmode)	{
				case 0:
				case 1:
				case 2:
				case 3:	{
					dst = (uint8*)(g_pnobridge->vram)+((((display_width-width)>>1)+((display_height-height-(genZ_renderingmode<2?BELTBAR_SIZE:0))>>1)*display_pitch)<<1);
					ljp_asmfastrender320(dst,src,height,(display_pitch-320)<<1);
					break;
				}
				case 4:					
					if (device_type&BF_ZODIAC){
						dst = ((uint8*) genZ_zodvram);
						ljp_asmfastrender320(dst,src,height,(480-320)<<1);
						if (genZ_smoothmode&_ZOD_SMOOTHING_) {
							stubzodiacblit(3,320,height,5);
							break;
						} else if (!genZ_smoothmode) {
							stubzodiacblit(3,320,height,1);
							break;
						}
					}
					if (device_type&BF_HIRESPLUS){
						dst = (uint8*)(g_pnobridge->vram);//+((((display_width-width)>>1)+((display_height-height)>>1)*display_pitch)<<1);
						if (genZ_smoothmode) fgbZ_filterrender(dst,src,(UInt32)((display_pitch)*2),512*2,SCALE_2XSAI,320,height,480,320/*-BELTBAR_SIZE*/);
					}
					break;
			}	
		  } else {
		    dst = ((UInt16*) g_pnobridge -> vram) + ( ((g_pnobridge -> display_height-120-BELTBAR_SIZE)>>1) * display_pitch ) + ((display_width-160)>>1);
			fgbZ_filterrender((uint8*)dst,(uint8*)src,display_pitch*2,512*2,SCALE_2XSAI,320,240,160,130);
		  }
		}		
	}
	
}

void genZ_MainLoop(void)
{
    update_input();
    
    genZ_frameBeginTime+=genZ_frameTime*100/genZ_speedpct;
    
    if ((genZ_frameskip==AUTO_FSKIP)&&(!genZ_fastforward))
	{
		int i,haswaited=0;
		
		//do avoid issue when Tick is looping
		if ((int32)(genZ_frameBeginTime-GetTickCount())>=(int32)(20*genZ_frameTime))
			genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;					
			
		while (GetTickCount()<=genZ_frameBeginTime) haswaited=1;
//		i=genZ_frameBeginTime-GetTickCount()+1;
//        if (i>0) {haswaited=1; SysTaskDelay(i);}
	
		if (haswaited||(genZ_frameskipped>=MAX_FRAME_SKIPPED)) 
		{
			genZ_drawframe=1;
			genZ_frameskipped=0;
		}
		else 
		{
			genZ_drawframe=0;
			genZ_frameskipped++;
		}
	}
	else
	{		
		if (genZ_frameskip==SKEEZIX_FSKIP) genZ_frameskipped=SKEEZIX_FSKIP;
		genZ_frameskipped++;
		
		if (genZ_fastforward)
		{
			if (genZ_frameskipped>GENZ_TURBOFSKIP)
			{
				genZ_frameskipped=0;
				genZ_drawframe=1;
			}
			else genZ_drawframe=0;
		}
		else
		{
			if (genZ_frameskipped>genZ_frameskip)
			{
				genZ_frameskipped=0;
				genZ_drawframe=1;				
			
			}
			else genZ_drawframe=0;
		}
		
		if ((!genZ_fastforward)&&(genZ_drawframe||(genZ_enabledSound&2)))
	    {	
			//do avoid issue when Tick is looping
			int i;
			if ((int32)(genZ_frameBeginTime-GetTickCount())>=(int32)(20*genZ_frameTime))
				genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
				
			while (GetTickCount()<=genZ_frameBeginTime) ;		
//			i=genZ_frameBeginTime-GetTickCount()+1;
//	        if (i>0) {SysTaskDelay(i);}
		}
	}


    if(!system_frame(!genZ_drawframe)) 
    {
    	menu_message("lock");
    	menu_waitPen(NULL,NULL,1);
    	system_reset();
    }
	genZ_updateScreen();
	
	
	genZ_framecount++;
}

void genZ_pause(int pause)
{	
   if (pause)
	{
		if (snd_stream) stub_SndStreamGetVolume(snd_stream,&snd_stream_vol);  
		if ((genZ_enabledSound&2)&&snd_stream)
		{
			stub_SndStreamStop(snd_stream);
			stub_SndStreamDelete(snd_stream);
			snd_stream=NULL;
		}
	}
	else
	{
		genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
		if (genZ_enabledSound&2) 
		{
			stub_SndStreamCreate(&snd_stream,sndOutput,genZ_sndfreq,sndInt16Little,sndStereo,genZ_sndCallBack,&reg_sys,genZ_sndbuffsize,1);
			stub_SndStreamStart(snd_stream);		
			stub_SndStreamSetVolume(snd_stream,snd_stream_vol);
	}	}
}


void genZ_changefskip(void)
{
	char tmp[64];
	int i;
	
	genZ_frameskip++;
	if (genZ_frameskip>SKEEZIX_FSKIP) genZ_frameskip=0;
	
	genZ_specialfskip=0;
	if (genZ_frameskip==SKEEZIX_FSKIP) genZ_specialfskip=1;
	
	if (genZ_frameskip==AUTO_FSKIP) sprintf(tmp,"Frameskip : AUTO");
	else if (genZ_frameskip==SKEEZIX_FSKIP) sprintf(tmp,"Frameskip : Skeezix");
	else sprintf(tmp,"Frameskip : %d",genZ_frameskip);	
	//DrawMessage(tmp,0);
	menu_message(tmp);
	
	i=GetTickCount();
	while ((GetTickCount()-i)<300)
	{
	}
	genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
}
void genZ_changerendering(void)
{	
	genZ_renderingmode++;
	if (genZ_renderingmode==5) genZ_renderingmode=0;
	
	if (genZ_renderingmode==2) {
		stub_fullscreen(1);
		LJP_HANDLE_EVENT(genZ_pause)	
	}
	if (genZ_renderingmode==0) {
		stub_fullscreen(0);
		LJP_HANDLE_EVENT(genZ_pause)
	}

	menu_message("Changing screen mode");
	LJP_WAIT(300)
	ClearScreen();
	genZ_frameBeginTime=GetTickCount()+genZ_frameTime*100/genZ_speedpct;
}


void genZ_updatevpad(unsigned int val)
{
	genZ_padState=val;
}

/***********************************************************/

int genZ_ShowMenuGFXConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  int scrmode,smoothing,fskipval,hwsmooth;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(19*sizeof(RectangleType));
  char tmp[16];
  

	
  scrmode=genZ_renderingmode;smoothing=genZ_smoothmode&_SMOOTH_MASK_;
  hwsmooth=genZ_smoothmode&_ZOD_SMOOTHING_;

  fskipval=genZ_frameskip;

	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {	  	
	    GUI_REDRAW_BG(0)
		
		menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  	
	  
	  	menu_button(10,80,(char*)"FRAMESKIP",0,&item_rect[1],2);	  
	  	menu_bar(150,80,fskipval,11,&item_rect[2]);
	  	if (fskipval<10) sprintf(tmp,"%d",fskipval);
	  	else if (fskipval==10) sprintf(tmp,"AUTO");
	  	else sprintf(tmp,"SKZX");
	  	menu_messageRaw(260,80,tmp);
	  
	  	menu_button(10,130,(char*)"ZOOM",0,&item_rect[3],2);
	  	menu_button(150,130,(char*)"1",((scrmode==0)||(scrmode==2)?2:1),&item_rect[4],0);
	  	menu_button(200,130,(char*)"1.25",((scrmode==1)||(scrmode==3)?2:1),&item_rect[5],0);
	  	menu_button(250,130,(char*)"WIDE",(scrmode==4?2:1),&item_rect[6],0);
	  	
	  	menu_button(10,180,(char*)"FULLSCREEN",0,&item_rect[7],2);
	  	menu_button(200,180,(char*)"OFF",(scrmode>=2?1:2),&item_rect[8],0);
	  	menu_button(250,180,(char*)"ON",(scrmode>=2?2:1),&item_rect[9],0);	  	  	  
	  
		menu_button(10,230,(char*)"SMOOTHING",0,&item_rect[10],2);
	  	menu_button(150,230,(char*)"-",((smoothing==0)&&(!hwsmooth)?2:1),&item_rect[11],4);
	  	menu_button(180,230,(char*)"1",(smoothing==1?2:1),&item_rect[12],4);
	  	menu_button(210,230,(char*)"2",(smoothing==2?2:1),&item_rect[13],4);
	  	menu_button(240,230,(char*)"3",(smoothing==3?2:1),&item_rect[14],4);
	  	menu_button(280,230,(char*)"HW",(hwsmooth?2:1),&item_rect[15],3);
	  	
	  	menu_button(10,280,(char*)"FLIPPING",0,&item_rect[16],2);
	  	menu_button(150,280,(char*)"OFF",(genZ_flipping?1:2),&item_rect[17],0);
	  	menu_button(200,280,(char*)"ON",(genZ_flipping?2:1),&item_rect[18],0);
	  	
	  	
	  	if (scr) ljp_WinScreenUnlock();
	
  	  }
	  redraw=menu_waitPen(&x,&y,0);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //FSKIP
	  if (TEST_PEN_MENU(x,y,item_rect[2])) {
	  	fskipval=(x-10-item_rect[2].topLeft.x)*11/100;
	  	if (fskipval<0) fskipval=0;
	  	if (fskipval>11) fskipval=11;
	  	redraw=2;
	  }
	  //SCREEN MODES
	  if (TEST_PEN_MENU(x,y,item_rect[4])) {
		  scrmode&=~4;
	  	  scrmode&=~1;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[5])) {
		  scrmode&=~4;
	  	  scrmode|=1;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[6])) {
	  	  scrmode=4;
	  	  redraw=1;
	  }
	  //SCREEN MODES FULL/WIN
	  if (TEST_PEN_MENU(x,y,item_rect[8])) {
	  	  scrmode&=~(2|4);
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) {
	  	  if (scrmode<2) scrmode|=2;
	  	  redraw=1;
	  }
	  //SMOOTHING
	  if (TEST_PEN_MENU(x,y,item_rect[11])&&scrmode) {
		  if (smoothing==2){DeInitStretchTables();}
	  	  smoothing=0;	  	  
	  	  hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[12])&&scrmode) {
		  if (smoothing==2){DeInitStretchTables();}
	  	  smoothing=1;
	  	  hwsmooth=0;
	  	  redraw=1;	  	  
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[13])&&scrmode) {
	  	  if (smoothing!=2){	 
			if (InitStretchTables()){
				menu_message("Err:InitStretchTables");
			}else{
		  	  smoothing=2;
		  	  hwsmooth=0;
		  	}		
	  	  }
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[14])&&scrmode) {
		  if (smoothing==2){DeInitStretchTables();}
	  	  smoothing=3;
	  	  hwsmooth=0;
	  	  redraw=1;	  	  
	  }
	  //HW SMOOTHING
	  if (TEST_PEN_MENU(x,y,item_rect[15])&&scrmode) {
	  	  hwsmooth=_ZOD_SMOOTHING_;
	  	  if (smoothing==2){DeInitStretchTables();}
	  	  smoothing=0;
	  	  redraw=1;	  	  
	  }
	  //FLIPPING
	  if (TEST_PEN_MENU(x,y,item_rect[17])) {
	  	  genZ_flipping=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[18])) {
	  	  genZ_flipping=1;
	  	  redraw=1;
	  }
	  
	  if (mStopEmulation) break;
  }  
  
  genZ_frameskip=fskipval;

  genZ_specialfskip=0;
  
  if (genZ_frameskip==SKEEZIX_FSKIP) 
  {	
    genZ_specialfskip=1;
  }
    
	genZ_renderingmode=scrmode;
	genZ_smoothmode=smoothing|hwsmooth;
  
  if (!mStopEmulation) menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}

int genZ_ShowMenuSNDConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  int vol;
  int bassboost;
  RectangleType *item_rect;
  char tmp[16];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  
  vol=snd_stream_vol/16;  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
	    GUI_REDRAW_BG(0)

		
		menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  
		menu_button(10,80,(char*)"SOUND",0,&item_rect[1],2);
		menu_button(150,80,(char*)"OFF",(genZ_enabledSound==0?2:1),&item_rect[2],0);
		menu_button(200,80,(char*)"SIM",(genZ_enabledSound&1?2:1),&item_rect[3],0);	  
		menu_button(250,80,(char*)"ON",(genZ_enabledSound&2?2:1),&item_rect[4],0);	 
		
		menu_button(150,120,(char*)"Z80",(genZ_enabledSound&4?2:1),&item_rect[5],0);	  
		menu_button(200,120,(char*)"FST",(genZ_enabledSound&8?2:1),&item_rect[6],0);	 
	  
		menu_button(10,160,(char*)"VOLUME",0,&item_rect[7],2);	  
	  	menu_bar(150,160,vol,128,&item_rect[8]);
	  	sprintf(tmp,"%d%",vol*100/64);	  	
	  	menu_messageRaw(270,160,tmp);
	  
		if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,0);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //SOUND OFF
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
	  	  if (genZ_enabledSound)
	  	  {
	  	  	if (snd.enabled) audio_shutdown();
		    snd.enabled=0;
			emul_z80=0;
	  	  }
		  genZ_enabledSound=0;		  
		  redraw=1;
	  }
	  //SOUND SIM
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
	  	  if (!(genZ_enabledSound&4)) 
	  	  {	
		  	 if (menu_confirm("RESET needed. Continue ?"))
	  	  	 {
	  	  	 	if (snd.enabled) audio_shutdown();
			    snd.enabled=0;
				emul_z80=1;
	  	  	 	genZ_enabledSound&=~3;
  		  	  	genZ_enabledSound|=1|4;
  		  	  	system_reset();
	  	  	 }
	  	  	
	  	  }
	  	  else
	  	  {
		  	  genZ_enabledSound&=~3;
  		  	  genZ_enabledSound|=1|4;
  		  }
	  	  redraw=1;
	  }
	  //SOUND ON
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
	  	  if (!(genZ_enabledSound&2))
	  	  {
	  	  	audio_init(genZ_sndfreq,genZ_sndbuffsize);
	  	  }	  	  
	  	  genZ_enabledSound&=~3;
		  genZ_enabledSound|=2;
		  redraw=1;
	  }
	  //ZX80 on/off
	  if (TEST_PEN_MENU(x,y,item_rect[5])) 
	  {
	  	if (!(genZ_enabledSound&4))
	  	{
	  		if (menu_confirm("RESET needed. Continue ?"))
	  	  	{	  	  	 	
				emul_z80=1;
  		  	  	genZ_enabledSound|=4;
  		  	  	system_reset();
	  	  	}	  	  	
	  	}
	  	else
	  	{
	  	 	emul_z80=0;
  		  	genZ_enabledSound&=~4;
	  	}
	  	/*wait*/
	  	int i=GetTickCount();
		while ((GetTickCount()-i)<300) {}
		redraw=1;
	  }
	  
	  //fast on/off
	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	    if (!(genZ_enabledSound&8))
	    {
		  	if (!(genZ_enabledSound&2))
		  	{
	  			audio_init(genZ_sndfreq,genZ_sndbuffsize);  		  		
	  			genZ_enabledSound&=~3;
	  			genZ_enabledSound|=2;
			}
		  	genZ_enabledSound|=8;		  	
		}
		else
		{
			genZ_enabledSound&=~8;
		}
		audio_fastmode(genZ_enabledSound&8);
		/*wait*/
	  	int i=GetTickCount();
	  	while ((GetTickCount()-i)<300) {}
		redraw=1;
	  }
	  
	  //VOLUME
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	vol=(x-item_rect[8].topLeft.x-10)*128/100;
	  	if (vol<0) vol=0;
	  	if (vol>128) vol=128;
	  	redraw=2;
	  }
	  if (mStopEmulation) break;
  }  
  
	snd_stream_vol=vol*16;  
	
  if (!mStopEmulation) menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}

int genZ_ShowMenuOTHERConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  char tmp[16];
  item_rect=(RectangleType*)ljz_malloc(36*sizeof(RectangleType));
  
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG(0)
		
  		
  	  	  	  	  	  	
	  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(10,50,(char*)"AUTO A",0,&item_rect[1],4);
	  menu_button(100,50,(char*)"OFF",(genZ_autofireA==0?2:1),&item_rect[2],0);
	  menu_button(150,50,(char*)"SLW",(genZ_autofireA==1?2:1),&item_rect[3],0);
	  menu_button(200,50,(char*)"MED",(genZ_autofireA==2?2:1),&item_rect[4],0);
	  menu_button(250,50,(char*)"FST",(genZ_autofireA==3?2:1),&item_rect[5],0);	  
	  menu_button(10,80,(char*)"AUTO B",0,&item_rect[6],4);
	  menu_button(100,80,(char*)"OFF",(genZ_autofireB==0?2:1),&item_rect[7],0);
	  menu_button(150,80,(char*)"SLW",(genZ_autofireB==1?2:1),&item_rect[8],0);
	  menu_button(200,80,(char*)"MED",(genZ_autofireB==2?2:1),&item_rect[9],0);
	  menu_button(250,80,(char*)"FST",(genZ_autofireB==3?2:1),&item_rect[10],0);
	  menu_button(10,110,(char*)"AUTO C",0,&item_rect[11],4);
	  menu_button(100,110,(char*)"OFF",(genZ_autofireC==0?2:1),&item_rect[12],0);
	  menu_button(150,110,(char*)"SLW",(genZ_autofireC==1?2:1),&item_rect[13],0);
	  menu_button(200,110,(char*)"MED",(genZ_autofireC==2?2:1),&item_rect[14],0);
	  menu_button(250,110,(char*)"FST",(genZ_autofireC==3?2:1),&item_rect[15],0);
	  
	  menu_button(10,250,(char*)"EMU SPEED",0,&item_rect[31],4);
	  menu_bar(130,250,genZ_speedpct,100,&item_rect[32]);
	  StrIToA(tmp,genZ_speedpct);
	  strcat(tmp,"%");
    	menu_messageRaw(240,250,tmp);
    	
      menu_button(10,285,(char*)"CPU CORE",0,&item_rect[33],4);
 	  menu_button(130,285,(char*)"C",(genZ_cpu_cyclone==0?2:1),&item_rect[34],0);
	  menu_button(180,285,(char*)"ASM",(genZ_cpu_cyclone==1?2:1),&item_rect[35],0);
	  
		if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,0);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //AUTOFIRE A
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
		  genZ_autofireA=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
		  genZ_autofireA=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
		  genZ_autofireA=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[5])) 
	  {
		  genZ_autofireA=3;
		  redraw=1;
	  }
	  //AUTOFIRE B
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
		  genZ_autofireB=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
		  genZ_autofireB=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
		  genZ_autofireB=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[10])) 
	  {
		  genZ_autofireB=3;
		  redraw=1;
	  }
	  //AUTOFIRE C
	  if (TEST_PEN_MENU(x,y,item_rect[12])) 
	  {
		  genZ_autofireC=0;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[13])) 
	  {
		  genZ_autofireC=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[14])) 
	  {
		  genZ_autofireC=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[15])) 
	  {
		  genZ_autofireC=3;
		  redraw=1;
	  }
	  //SPEED
	  if (TEST_PEN_MENU(x,y,item_rect[32])) 
	  {
	  	genZ_speedpct=(x-10-item_rect[32].topLeft.x);
	  	if (genZ_speedpct>100) genZ_speedpct=100;
	  	if (genZ_speedpct<10) genZ_speedpct=10;
	  	redraw=2;
	  }
	  //CPU CORE
	  if (TEST_PEN_MENU(x,y,item_rect[34])) 
	  {
		  if (menu_confirm("RESET needed. Continue ?")) 
		  {
			genZ_cpu_cyclone=0;		  
		  	system_reset();
		  }
		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[35])) 
	  {
		  if (menu_confirm("RESET needed. Continue ?")) 
		  {
			genZ_cpu_cyclone=1;
		  	system_reset();
		  }
		  redraw=1;
	  }
	  if (mStopEmulation) break;
  }  
  
  if (!mStopEmulation) menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}




int genZ_ShowMenuSave(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  
  
  FILE *savefile;
  int slot[4];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  memset((char*)item_rect,0,15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
  	  	GUI_REDRAW_BG(0)
		
  	  	  	  	  	  	
	 
  	  slot[0]=slot[1]=slot[2]=slot[3]=0;
  	  
  	  

	  
	  for (int i=0;i<4;i++)	  
	  if (!state_load(i,1))
	  {	  
	  		UInt32 long_time;
			DateTimeType date_time;		
			char str[255];
			
			slot[i]=1;
			
			GetStateFileName(i,str);
			savefile=fopen(str,"rb");
		    if (savefile) 
	  		{				  	
			  	fclose(savefile);	
			  	fgetdate(str,&long_time);		
				TimSecondsToDateTime(long_time,&date_time);
				//date is an ofs from 01 Jan 1900 00:00:00			
				sprintf(str,"%02d%s%02d-%02d:%02d",
				date_time.day,str_month[date_time.month-1],date_time.year%100,date_time.hour,date_time.minute);
				if (i) 
				{
					menu_messageRaw(150,80+i*30,str);
					menu_button(10,80+i*30,"X",5,&item_rect[6+i],3);
				}
				else 
				{
					menu_messageRaw(150,200,str);
					menu_button(10,200,"X",5,&item_rect[6+i],3);
				}
		    }			 
			
	  	}	
  	  	  	  	  	  	
	menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
	menu_button(10,50,(char*)"SAVE",0,&item_rect[1],2);	  	  	  
	menu_button(50,110,(char*)"SLOT 1",(slot[1]?2:3),&item_rect[2],1);
	menu_button(50,140,(char*)"SLOT 2",(slot[2]?2:3),&item_rect[3],1);
	menu_button(50,170,(char*)"SLOT 3",(slot[3]?2:3),&item_rect[4],1);
	menu_button(50,200,(char*)"QUICK",(slot[0]?2:3),&item_rect[5],1);
	  
	  
	  
		if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	if (menu_confirm("Delete QUICKSLOT?"))
	  	{
		  	char str[255];		
			GetStateFileName(0,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	if (menu_confirm("Delete SLOT 1?"))
	  	{
		  	char str[255];		
			GetStateFileName(1,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	if (menu_confirm("Delete SLOT 2?"))
	  	{
		  	char str[255];		
			GetStateFileName(2,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	if (menu_confirm("Delete SLOT 3?"))
	  	{
		  	char str[255];		
			GetStateFileName(3,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //SAVE 0
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
	  	  if (!slot[1]||(SHOW_OVERWRITE))
		  {state_save(1);result=1;break;}
		  else redraw=1;
		  
	  }
	  //SAVE 1
	  if (TEST_PEN_MENU(x,y,item_rect[3]))
	  {
		  if (!slot[2]||(SHOW_OVERWRITE))
		  {state_save(2);result=1;break;}
		  else redraw=1;
		
	  }
	  //SAVE 2
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
    	  if (!slot[3]||(SHOW_OVERWRITE))
		  {state_save(3);result=1;break;}
		  else redraw=1;
		
	  }
	  //QUICKSAVE
	  if (TEST_PEN_MENU(x,y,item_rect[5]))
	  {
	  	  if (!slot[0]||(SHOW_OVERWRITE))
		  {state_save(0);result=1;break;}
		  else redraw=1;
		
	  }
	  
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}



int genZ_ShowMenuLoad(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  FILE *savefile;
  int slot[4];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  memset((char*)item_rect,0,15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG(0)	  	  	    	  
		slot[0]=slot[1]=slot[2]=slot[3]=0;  	    	  
		for (int i=0;i<4;i++)	  
  	  	if (!state_load(i,1)) 
	  	{	  
	  		UInt32 long_time;
			DateTimeType date_time;		
			char str[255];
			
			slot[i]=1;
			GetStateFileName(i,str);
			savefile=fopen(str,"rb");
		    if (savefile) 
	  		{	
			  	
			  	fclose(savefile);	
			  	fgetdate(str,&long_time);		
				TimSecondsToDateTime(long_time,&date_time);
				//date is an ofs from 01 Jan 1900 00:00:00			
				sprintf(str,"%02d%s%02d-%02d:%02d",
				date_time.day,str_month[date_time.month-1],date_time.year%100,date_time.hour,date_time.minute);
				if (i) 
				{
					menu_messageRaw(150,80+i*30,str);
					menu_button(10,80+i*30,"X",5,&item_rect[6+i],3);					
				}
				else 
				{
					menu_messageRaw(150,200,str);
					menu_button(10,200,"X",5,&item_rect[6+i],3);
				}
		    }
	  	}
  	  	  	  	  	  	
	  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
	menu_button(10,50,(char*)"LOAD",0,&item_rect[1],2);	  	  	  
	menu_button(50,110,(char*)"SLOT 1",(slot[1]?2:3),&item_rect[2],1);
	menu_button(50,140,(char*)"SLOT 2",(slot[2]?2:3),&item_rect[3],1);
	menu_button(50,170,(char*)"SLOT 3",(slot[3]?2:3),&item_rect[4],1);
	menu_button(50,200,(char*)"QUICK",(slot[0]?2:3),&item_rect[5],1);

		if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;

	  
 	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	if (menu_confirm("Delete QUICKSLOT?"))
	  	{
		  	char str[255];		
			GetStateFileName(0,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	if (menu_confirm("Delete SLOT 1?"))
	  	{
		  	char str[255];		
			GetStateFileName(1,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	if (menu_confirm("Delete SLOT 2?"))
	  	{
		  	char str[255];		
			GetStateFileName(2,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	if (menu_confirm("Delete SLOT 3?"))
	  	{
		  	char str[255];		
			GetStateFileName(3,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //LOAD 0
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  if (slot[1])
		if (!state_load(1,0)) {result=1;break;}
	  }
	  //LOAD 1
	  if (TEST_PEN_MENU(x,y,item_rect[3]))
	  {
		  if (slot[2])
		if (!state_load(2,0)) {result=1;break;}
	  }
	  //LOAD 2
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (slot[3])
		if (!state_load(3,0)) {result=1;break;}	  
	  }
	
	  //QUICKLOAD
	  if (TEST_PEN_MENU(x,y,item_rect[5]))
	  {
	  	if (slot[0])
    	if (!state_load(0,0)) {result=1;break;}
	  }
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}


int genZ_ShowMenuConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG(0)
		
		
		menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  
	  	menu_button(92,100,(char*)"GFX",0,&item_rect[1],2);
	  	menu_button(92,150,(char*)"SOUND",0,&item_rect[2],2);
	  	menu_button(92,200,(char*)"OTHER",0,&item_rect[3],2);	  
		if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;

	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //GFX
	  if (TEST_PEN_MENU(x,y,item_rect[1])) 
	  {
	  	genZ_ShowMenuGFXConfig(savescr);
		redraw=1;
	  }
	  //SOUND
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
     	genZ_ShowMenuSNDConfig(savescr);
		redraw=1;
	  }
	  //OTHER
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {	  	  		  
		  genZ_ShowMenuOTHERConfig(savescr);
		  redraw=1;
	  }
	  if (mStopEmulation) break;

  }  
  ljz_free(item_rect);
  return result;
}


int genZ_ShowSysInfos(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect,my_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG(0)
		
  	  	
				
		char str[64],strb[32];
		unsigned long maxP,freeP,usedP;		
		VolumeInfoType vinfo;
		VFSVolumeSize(genZ_curVolume,&usedP,&maxP);		
		VFSVolumeGetLabel(genZ_curVolume,strb,16);strb[16]=0;
		menu_button(10,70,"Volume",0,&my_rect,4);		
		sprintf(str,"%s[%d]",strb,genZ_curVolume);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,63,str);				
		sprintf(str,"Used:%dM/%dM",usedP>>20,maxP>>20);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,81,str);				
		MemHeapFreeBytes(0,&freeP,&maxP);		
		menu_button(10,100,"RAM",6,&my_rect,4);		
		sprintf(str,"RAM %ldK/%ldK",freeP/1024,maxP/1024);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,100,str);		
		
		

  	
	  	menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
		
	  
		if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;

	  
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}




int genZ_ShowMenu()
{
  int x,y;
  int redraw;
  int result;
  int slot0;
  uint8 *scr,*savescr,*tmpscr,*tmpdst;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));  
  redraw=1;
 

  genZ_pause(1);
    
       
  scr=/*(uint8*)(g_pnobridge -> vram);*/ljp_WinScreenLock(winLockCopy);  
  savescr=(uint8*)ljz_malloc(display_width*display_height*2);    
  if (scr&&savescr)
  {
	  tmpscr=scr;
  	for (int yy=0;yy<display_height-(genZ_renderingmode<2?BELTBAR_SIZE:0);yy++)
  	{
	  for (x=0;x<display_width;x++)
	  {
  		y=((uint16*)tmpscr)[x];
	  	((uint16*)tmpscr)[x]=((y>>12)<<11)|(((y>>6)&31)<<5)|((y>>1)&15);
	  }
	  tmpscr+=display_pitch*2;
	}
	tmpscr=scr;
	tmpdst=savescr;
	for (int yy=0;yy<display_height;yy++)
  	{
	    memcpy(tmpdst,tmpscr,display_width*2);
	    tmpdst+=display_width*2;
	    tmpscr+=display_pitch*2;
	}
  }
		if (scr) ljp_WinScreenUnlock();
            
              
  for (;;)
  {  
  	if (redraw)
  	{  	
	  	slot0=!state_load(0,1);    
	  	
		GUI_REDRAW_BG(1)

  		  		  		
		menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
		menu_button(10,100,(char*)"SAVE",0,&item_rect[1],2);
		menu_button(180,100,(char*)"QUICK SAVE",(slot0?2:1),&item_rect[2],2);
		menu_button(10,150,(char*)"LOAD",0,&item_rect[3],2);
		menu_button(180,150,(char*)"QUICK LOAD",(slot0?2:3),&item_rect[4],2);
		menu_button(10,200,(char*)"CONFIG",0,&item_rect[5],2);
		menu_button(180,250,(char*)"LAUNCHER",4,&item_rect[6],2);
		
		menu_button(180,200,(char*)"SYS INFOS",7,&item_rect[8],2);
		
		menu_button(10,250,(char*)"RESET GEN.",9,&item_rect[7],2);				
		if (scr) ljp_WinScreenUnlock();
		
  	}
	  redraw=menu_waitPen(&x,&y,1);
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;

	  //HOME
	  if (TEST_PEN_MENU(x,y,item_rect[6])) {result=1;break;}
	  //BACK
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //CONFIG
	  if (TEST_PEN_MENU(x,y,item_rect[5])) {genZ_ShowMenuConfig(savescr);redraw=1;}
	  //LOAD
	  if (TEST_PEN_MENU(x,y,item_rect[3])) {if (genZ_ShowMenuLoad(savescr)) {result=0;break;} redraw=1;}
	  //SAVE
	  if (TEST_PEN_MENU(x,y,item_rect[1])) {if (genZ_ShowMenuSave(savescr)) {result=0;break;} redraw=1;}
	  //QUICK SAVE
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  state_save(0);result=0;break;
	  }
	  //QUICK LOAD
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (slot0)
		    if (!state_load(0,0)) {result=0;break;}
	  }
	  //RESET GB
	  if (TEST_PEN_MENU(x,y,item_rect[7])) {if (menu_confirm("Reset GENESIS ?")) {system_reset(); result=0;break;}redraw=1;}
	  
	  //SYS INFOS
	  if (TEST_PEN_MENU(x,y,item_rect[8])) {genZ_ShowSysInfos(savescr);redraw=1;}

		if (mStopEmulation) break;
	}
  ljz_free(item_rect);
  
  if (savescr) ljz_free(savescr);
            
  genZ_pause(0);
  return result;
}  
