//---------------------------------------------------------------------------
// NEOPOP : Emulator as in Dreamland
//
// Copyright (c) 2001-2002 by neopop_uk
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version. See also the license.txt file for
//	additional informations.
//---------------------------------------------------------------------------

/*
//---------------------------------------------------------------------------
//=========================================================================

	system_main.c

//=========================================================================
//---------------------------------------------------------------------------

  History of changes:
  ===================

//---------------------------------------------------------------------------
*/

#include "neopop.h"
#include <math.h>			//ceil

#include "system_main.h"
#include "system_graphics.h"
#include "system_input.h"
#include "system_sound.h"
#include "system_rom.h"

#include "filters.h"
//#include "system_language.h"
//#include "system_comms.h"


//_u8 system_frameskip_key;
//extern _u8 *bios;
#include "PenInputMgr.h"
//=============================================================================

extern UInt32 device_type;
extern UInt32 ljp_isLowRes,ljp_confirmOW,ljp_confirmSL;

extern int mStopEmulation;
UInt16 *neopopZ_zodvram;

char str_month[12][4]=
{"Jan","Feb","Mar","Apr","May","Jun",
"Jul","Aug","Sep","Oct","Nov","Dec"};

char shortrom_filename[256];
char RomPath[256];
char neopopZ_curVolume;

int neopopZ_loadrom();
void neopopZ_end();

void neopopZ_changefskip(void);
void neopopZ_changerendering(void);
void neopopZ_updatevpad(unsigned int val);
void neopopZ_MainLoop(void);
void neopopZ_pause(int pause);
int neopopZ_ShowMenu(void);
void neopopZ_updateScreen(void);


_u32 neopopZ_enabledSound,neopopZ_sndfreq;
_u32 neopopZ_renderingmode,neopopZ_smoothmode,neopopZ_flipping;
_u32 neopopZ_padState;
_s32 neopopZ_frameskipped,neopopZ_frameskip,neopopZ_fastforward;
_s32 neopopZ_frameBeginTime;
_s32 neopopZ_speedpct;
_s32 neopopZ_autofireA,neopopZ_autofireB;
_s32 neopopZ_framecount,neopopZ_drawframe;
_u32 neopopZ_emu_accuracy;

_s32 neopopZ_frameTime;

 


_u16 *neopopZ_ColorConv;
_u16 *vidram;

BOOL paused = FALSE;

//=============================================================================

#define FRAMERATECOUNT	60	//Number of frames between updates.

//=============================================================================

#include "SoundMgr.h"
extern SndStreamRef snd_stream;
uint32 snd_stream_vol;

char 	savename[255];

void emu_resettimer(void) {
	neopopZ_frameBeginTime=GetTickCount()+17*100/neopopZ_speedpct;
}

void neopopZ_togglesound(void) {
	char tmp[64];
	int i;
	
	neopopZ_pause(1);
	
	neopopZ_enabledSound++;
	if (neopopZ_enabledSound>2) neopopZ_enabledSound=0;
	
	if (neopopZ_enabledSound==0) strcpy(tmp,"Sound : OFF");
	else if (neopopZ_enabledSound==1) strcpy(tmp,"Sound : SIM");
	else if (neopopZ_enabledSound==2) strcpy(tmp,"Sound : ON");

	menu_message(tmp);
	
	LJP_WAIT(300);
	emu_resettimer();
	
	neopopZ_pause(0);
}

char neopopZ_reset(void) {
	neopopZ_pause(1);
	
  	if (menu_confirm("Reset NGP ?")) reset();
  	
  	emu_resettimer();
  	
  	neopopZ_pause(0);
  	
  	return 0;
}

void neopopZ_quicksave(void) {
	neopopZ_pause(1);
	
	if (CONFIRM_SAVE) state_save(0);
	emu_resettimer();
	ClearScreen();
	
	neopopZ_pause(0);
}

void neopopZ_quickload(void) {
	neopopZ_pause(1);
	
	if (!state_load(0,1))
	{
		if (CONFIRM_LOAD) state_load(0,0);
		emu_resettimer();
		ClearScreen();
	}
	
	neopopZ_pause(0);
}

void getstate_filename(int num,char *filename)
{
	char result[256];
	_u32 i=0,j=255,k=255,l=255;
	i=0;
	j=255;
	k=-1;
	while (RomPath[i]!=0) 
	{
		if (RomPath[i]=='.') j=i;
		if (RomPath[i]=='/') k=i;
		if (RomPath[i]=='\\') k=i;
		i++;
	  }
	if (j==255) j=i;  //pas d'extension	  
	memcpy(result,RomPath+k+1,j+1-k-1);    
	result[j+1-k-1]=0;
	strcat(result,"ns");
	sprintf(filename,"%s%s%d",NEOPOPZ_SAVESDIR,result,num);
	filename[0]=neopopZ_curVolume+'0';
}

int state_load(int num,int only_test)
{
	FILE	*file;	
	
	
	getstate_filename(num,savename);
	file = fopen(savename, "rb");
	if (file == NULL)
	{
		return -1;
	}
	fclose(file);
	if (only_test) return 0;
	
	
	menu_message("Loading");
	state_restore(savename);	

	return 0;
}

int state_save(int num)
{			
	getstate_filename(num,savename);	
	
	menu_message("Saving");
	
	state_store(savename);	
	return 0;
}


int init_video(void)
{
	//Init vid
					
    /*input*/
    vidram = (_u16*)ljz_malloc(SCREEN_HEIGHT*SCREEN_WIDTH*2);//160*152*2
    
    neopopZ_ColorConv = (_u16*)ljz_malloc(65536*2);
    int r,g,b;
    for (int i=0;i<65536;i++)
    {    	
    	r = ((i & 0x00F) << 12);
    	g = ((i & 0x0F0) << 3);
		b = ((i & 0xF00) >> 7);
		neopopZ_ColorConv[i]=r|g|b;
	}
	return 0;
}

void neopopZ_initparam(void)
{
	signed long val1,val2,val;
	char *ftrptr;
	FtrGet(CREATORID,FTR_OPT_SOUND_ID,(unsigned long*)&val);  
	neopopZ_enabledSound=val;
	FtrGet(CREATORID,FTR_OPT_SOUNDFREQ_ID,(unsigned long*)&val);  
	if (val==1) neopopZ_sndfreq=22050;
	else if (val==2) neopopZ_sndfreq=11025;
	else neopopZ_sndfreq=44100;	
	
	FtrGet(CREATORID,FTR_OPT_EMU_ACCURACY_ID,(unsigned long*)&val);  
	neopopZ_emu_accuracy=val;
	
	FtrGet(CREATORID,FTR_OPT_SCREENMODE_ID,(unsigned long*)&val1);
	FtrGet(CREATORID,FTR_OPT_SMOOTHING_ID,(unsigned long*)&val2);
	neopopZ_renderingmode=val1;
	neopopZ_smoothmode=val2;
		

	neopopZ_fastforward=0;	
	neopopZ_frameskipped=0;
	
    FtrGet(CREATORID,FTR_OPT_FSKIP_ID,(unsigned long*)&val); 
	neopopZ_frameskip=val;
	
	FtrGet(CREATORID,FTR_OPT_SOUNDVOLUME_ID,(unsigned long*)&val); 
	snd_stream_vol=val;
	if (snd_stream_vol<0) snd_stream_vol=0;
	if (snd_stream_vol>128) snd_stream_vol=128;
	snd_stream_vol*=16;  
	
	
	filters_init();	    



//Core settings
	language_english = TRUE;	


	
	
	neopopZ_speedpct=100;
	neopopZ_autofireA=0;
	neopopZ_autofireB=0;	
	neopopZ_framecount=0;
	neopopZ_drawframe=1;
	init_video();	
    

    strcpy(RomPath,NEOPOPZ_ROMSDIR	);
    
    
    {//handle categories
    	char cat[32];
	  	FtrGet(CREATORID,FTR_OPT_CATEGORY_ID ,(unsigned long*)&ftrptr);
	  	strcpy(cat,ftrptr);
  		if ( StrCaselessCompare(cat,"uNaRcHiVeD") != 0) {
  			strcat(RomPath,cat);
	  		strcat(RomPath,"/");
	  	}
	  	//FtrPtrFree(CREATORID, FTR_OPT_CATEGORY_ID);
  	}
  	            
    FtrGet(CREATORID,FTR_ROM_FILENAME_ID ,(unsigned long*)&ftrptr);
    
    neopopZ_curVolume=ftrptr[0];
    RomPath[0]=neopopZ_curVolume+'0';
    
    ftrptr+=3;
    strcpy(shortrom_filename,ftrptr);  
    strcat(RomPath,ftrptr);
//    FtrPtrFree(CREATORID, FTR_ROM_FILENAME_ID);	

    
    neopopZ_frameTime=17; //hardcoded to NTSC
}

void neopopZ_audio_init()
{
	system_sound_init();
}

//extern _u8 *bios;
void neopopZ_initSystem(void)
{     	
     // a ce moment la on lance l'initialisation de l'emulateur
     // 1 - le bios	
//    bios= ljz_malloc(0x10000);   
//menu_confirm("bios install");
//bios=(_u8*)ljz_malloc(0x10000);
    bios_install();     	     
    // 3 - des variables     
    system_colour=COLOURMODE_AUTO;
    // 4 - reset de la machine
//menu_confirm("reset");
    reset();
//menu_confirm("init adr trans tab");
    init_adr_trans_tab();
    if (neopopZ_enabledSound==2) mute=FALSE;
    else mute=TRUE;
    
    
	paused = FALSE;
}
	
int neopopZ_loadrom()
{	
ljz_std_showstat();
	char strload[256];
	neopopZ_initparam();    			
	
	/* Load game */
    sprintf(strload,"Loading : %s",shortrom_filename);
    menu_message(strload);          		
//	unsigned long freeM,maxM;	
//	MemHeapFreeBytes(0,&freeM,&maxM);
//	StrPrintF(strload,"mem : %d %d",freeM,maxM);
//	menu_message(strload);
//	menu_waitPen(NULL,NULL,1);

    rom.data=NULL;    
    if(!system_load_rom(RomPath))
	{
		menu_message("Error loading file");
		menu_waitPen(NULL,NULL,1);
		return 1;
	}	
		
 // la rom est charge. ^_^
 
//menu_confirm("initSystem");
	neopopZ_initSystem();	
//menu_confirm("audio_init");
    neopopZ_audio_init();            

    neopopZ_frameBeginTime=GetTickCount();

	neopopZ_pause(0);
    return 0;  
}
void neopopZ_end()
{
	int val;

	
	
	neopopZ_pause(1);
	
	ljz_free(neopopZ_ColorConv);

	
/*emu*/

	
	  
/*	  if (saveram) 
	  {
	  	save_sram();
	  	ljz_free(saveram);
	  }*/

	system_unload_rom();
//	ljz_free(bios);
	/********/
      

    FtrSet(CREATORID,FTR_OPT_EMU_ACCURACY_ID,neopopZ_emu_accuracy);  
    
    FtrSet(CREATORID, FTR_OPT_SMOOTHING_ID,neopopZ_smoothmode);	
	FtrSet(CREATORID, FTR_OPT_SCREENMODE_ID,neopopZ_renderingmode);	

	FtrSet(CREATORID, FTR_OPT_SOUND_ID,neopopZ_enabledSound);
	switch (neopopZ_sndfreq)
	{
		case 11025:val=2;break;
		case 22050:val=1;break;
		default:val=0;break;
	}
	FtrSet(CREATORID, FTR_OPT_SOUNDFREQ_ID,val);
		
	FtrSet(CREATORID, FTR_OPT_FSKIP_ID,neopopZ_frameskip);
	
	FtrSet(CREATORID,FTR_OPT_SOUNDVOLUME_ID,snd_stream_vol/16); 
	
	ljz_free(vidram);
	
	ljz_std_showstat();

}

void neopopZ_MainLoop(void)
{
	//if(frameskip_count) frameskip_count--;
	//frameskip_count = neopopZ_drawframe;

	if (!neopopZ_enabledSound) emulate_ns(neopopZ_emu_accuracy);
	else emulate(neopopZ_emu_accuracy);
	
	//frameskipping
	//if (frameskip_count==0) frameskip_count==neopopZ_frameskip;
}





void neopopZ_changefskip(void)
{
	char tmp[64];
	int i;
	

	neopopZ_frameskip++;
	if (neopopZ_frameskip>SKEEZIX_FSKIP) neopopZ_frameskip=0;	
	
	if (neopopZ_frameskip==AUTO_FSKIP) sprintf(tmp,"Frameskip : AUTO");
	else if (neopopZ_frameskip==SKEEZIX_FSKIP) sprintf(tmp,"Frameskip : Skeezix");
	else sprintf(tmp,"Frameskip : %d",neopopZ_frameskip);	
	//DrawMessage(tmp,0);
	menu_message(tmp);
	
	i=GetTickCount();
	while ((GetTickCount()-i)<300)
	{
	}
	neopopZ_frameBeginTime=GetTickCount()+neopopZ_frameTime*100/neopopZ_speedpct;
}
void neopopZ_changerendering(void)
{	
	neopopZ_renderingmode++;
	if (neopopZ_renderingmode>3) neopopZ_renderingmode=0;
	if (neopopZ_renderingmode==2) {
		stub_fullscreen(1);
		LJP_HANDLE_EVENT(neopopZ_pause)	
	}
	if (neopopZ_renderingmode==0) {
		stub_fullscreen(0);
		LJP_HANDLE_EVENT(neopopZ_pause)
	}

	menu_message("Changing screen mode");
	neopopZ_frameBeginTime=GetTickCount();
	while ((GetTickCount()-neopopZ_frameBeginTime)<300)	{
	}
	neopopZ_frameBeginTime=GetTickCount()+neopopZ_frameTime*100/neopopZ_speedpct;
}


void neopopZ_updatevpad(unsigned int val)
{
	neopopZ_padState=val;
}

extern void neopopZ_fastrender(_u16 *src,_u16 *dest,_u16 *colorconv);

void neopopZ_asmfastrender(uint8 *screen,uint8 *source,uint32 offset_pitch)
{
	__asm{
		stmfd r13!,{r3-r12}

		mov	  r3,#152
		bic	  r0,r0,#3
		bic	  r1,r1,#3

	 loopY:	
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r4-r12}
		stmia r0!,{r4-r12}
		
		ldmia r1!,{r5-r12}
		stmia r0!,{r5-r12}
		
		add	  r0,r0,r2 //(256-160) * 2
		
		subs  r3,r3,#1
		bne   loopY
		
		ldmfd r13!,{r3-r12}	
		}	
}

void neopopZ_asmfastrenderx2(uint8 *screen,uint8 *source,uint32 offset_pitch)
{
	__asm{
	
#define asm_mac() \
		mov	  r5,r4,lsr #16;\
		mov	  r4,r4,lsl #16;\
		orr	  r5,r5,r5,lsl #16;\
		orr	  r4,r4,r4,lsr #16;\
		mov	  r7,r6,lsr #16;\
		mov	  r6,r6,lsl #16;\
		orr	  r7,r7,r7,lsl #16;\
		orr	  r6,r6,r6,lsr #16;\
		mov	  r9,r8,lsr #16;\
		mov	  r8,r8,lsl #16;\
		orr	  r9,r9,r9,lsl #16;\
		orr	  r8,r8,r8,lsr #16;\
		mov	  r11,r10,lsr #16;\
		mov	  r10,r10,lsl #16;\
		orr	  r11,r11,r11,lsl #16;\
		orr	  r10,r10,r10,lsr #16;	
	
		stmfd r13!,{r3-r12}
	
		mov	  r3,#152
		bic	  r0,r0,#3
		bic	  r1,r1,#3
		add	  r12,r0,r2
		mov	  r2,r2,lsl #1
		sub	  r2,r2,#640
	 loopY:	
		ldmia r1!,{r4,r6,r8,r10}  //read 4*4/2=8pixels				
		asm_mac()		
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		
#undef asm_mac		
		
		add	  r0,r0,r2 //(256-160) * 2
		add	  r12,r12,r2 //(256-160) * 2
		
		subs  r3,r3,#1
		bne   loopY
		
		ldmfd r13!,{r3-r12}	
		}	
}

void neopopZ_asmfastrenderx2_2(uint8 *screen,uint8 *source,UInt32 offset_pitch)
{
	__asm{
	
#define asm_mac() \
		mov	  r5,r4,lsr #16;\
		mov	  r4,r4,lsl #16;\
		orr	  r5,r5,r5,lsl #16;\
		orr	  r4,r4,r4,lsr #16;\
		mov	  r7,r6,lsr #16;\
		mov	  r6,r6,lsl #16;\
		orr	  r7,r7,r7,lsl #16;\
		orr	  r6,r6,r6,lsr #16;\
		mov	  r9,r8,lsr #16;\
		mov	  r8,r8,lsl #16;\
		orr	  r9,r9,r9,lsl #16;\
		orr	  r8,r8,r8,lsr #16;\
		mov	  r11,r10,lsr #16;\
		mov	  r10,r10,lsl #16;\
		orr	  r11,r11,r11,lsl #16;\
		orr	  r10,r10,r10,lsr #16;	
	
		stmfd r13!,{r3-r12}
	
		mov	  r3,#144
		bic	  r0,r0,#3
		bic	  r1,r1,#3
		add	  r12,r0,r2
		mov	  r2,r2,lsl #1
		sub	  r2,r2,#640
	 loopY:	
		ldmia r1!,{r4,r6,r8,r10}  //read 4*4/2=8pixels				
		asm_mac()		
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		ldmia r1!,{r4,r6,r8,r10}  
		asm_mac()
		stmia r0!,{r4-r11}
		stmia r12!,{r4-r11}
		
#undef asm_mac		
		
		add	  r0,r0,r2 //(256-160) * 2
		add	  r12,r12,r2 //(256-160) * 2
		
		subs  r3,r3,#1
		bne   loopY
		
		ldmfd r13!,{r3-r12}	
		}	
}

void neopopZ_updateScreen(void)
{	 
	if (1) //(frameskip_count==0)
	{
		int width = SCREEN_WIDTH;//(reg[12] & 1) ? 320 : 256;
	    int height = SCREEN_HEIGHT;//(reg[1] & 8) ? 240 : 224;
	    UInt16 *fline, *tline;
		UInt16 *f, *t, *tline2, *t2;
		int x,y;
		
		fline = (UInt16*)vidram;
	    
    	//if (width==256) neopopZ_asmfastrender256((_u8*)neopopZ_curRenderingScreenPtr,(_u8*)VideoBuffer,height,bitmap.viewport.x<<1);
    	//if (width==320)	neopopZ_asmfastrender320((_u8*)neopopZ_curRenderingScreenPtr,(_u8*)VideoBuffer,height,bitmap.viewport.x<<1);
    	//neopopZ_fastrender(src,dest,neopopZ_ColorConv);

    	uint16 *src=(uint16*)cfb;
    	uint16 *dest=(UInt16*)vidram;
    	//int r,g,b,col;    	
    	for (int y = 0; y < height; y++){
			for (int x = 0; x < width; x++)
				*dest++=neopopZ_ColorConv[*src++];			
		}
	  if (!ljp_isLowRes) {
		switch (neopopZ_renderingmode)
	  	{
	  		case 3:
	  		case 1:
	  			int offs=0;
	  			if ( (ljp_showBeltbar) && (display_height==320) ) {
	  				fline+= SCREEN_WIDTH * 4;
	  				tline = ((UInt16*) g_pnobridge -> vram) + /*0 +*/ ((display_width-SCREEN_WIDTH*2)>>1);
	  				offs=8;
	  			} else {
	  				tline = ((UInt16*) g_pnobridge -> vram) + ( ((g_pnobridge -> display_height-SCREEN_HEIGHT*2-(neopopZ_renderingmode<2?BELTBAR_SIZE:0))>>1) * display_pitch ) + ((display_width-SCREEN_WIDTH*2)>>1);
	  			}
	  			switch (neopopZ_smoothmode&_SMOOTH_MASK_){	  			
		  			case 7:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,SCALE_2XSAI,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,(SCREEN_HEIGHT-offs)*2);
		  			break;
		  			case 6:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,SUPER_EAGLE,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,SCREEN_HEIGHT*2);
		  			break;
			  		case 5:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,SUPER_2XSAI,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,SCREEN_HEIGHT*2);
		  			break;
				  	case 4:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,_2XSAI,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,SCREEN_HEIGHT*2);	  			
	  				break;
				  	case 3:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,HQ2X,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,SCREEN_HEIGHT*2);
		  			break;	
		  			case 2:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,LQ2X,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,SCREEN_HEIGHT*2);
		  			break;
			  		case 1:fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,ADVMAME2X,SCREEN_WIDTH,SCREEN_HEIGHT-offs,SCREEN_WIDTH*2,SCREEN_HEIGHT*2);
		  			break;
			  		case 0:
			  		
			  			if (device_type&BF_ZODIAC){	
				  			if (neopopZ_smoothmode&_ZOD_SMOOTHING_) {
							tline = ((UInt16*) neopopZ_zodvram);				
							neopopZ_asmfastrender((uint8*)tline,(uint8*)fline,(480-SCREEN_WIDTH)<<1);
							stubzodiacblit(3,SCREEN_WIDTH,SCREEN_HEIGHT,6);
							break;
						}
						else if (!neopopZ_smoothmode) {
							tline = ((UInt16*) neopopZ_zodvram);				
							neopopZ_asmfastrender((uint8*)tline,(uint8*)fline,(480-SCREEN_WIDTH)<<1);
							stubzodiacblit(3,SCREEN_WIDTH,SCREEN_HEIGHT,2);
							break;
						}
			  		}
			  		else {
			  			if ( (ljp_showBeltbar) && (display_height==320) ) {
			  				neopopZ_asmfastrenderx2_2((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1));
			  			} else {
			  				neopopZ_asmfastrenderx2((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1));
			  			}
						break;
					}
				}
				break;
			case 2:
	  		case 0:
				tline = ((UInt16*) g_pnobridge -> vram) + ( ((g_pnobridge -> display_height-SCREEN_HEIGHT-(neopopZ_renderingmode<2?BELTBAR_SIZE:0))>>1) * display_pitch ) + ((display_width-SCREEN_WIDTH)>>1);
			    neopopZ_asmfastrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch-SCREEN_WIDTH)<<1));
			    break;
			case 4:			
				if (device_type&BF_ZODIAC){					
					if (neopopZ_smoothmode&_ZOD_SMOOTHING_) {
						tline = ((UInt16*) neopopZ_zodvram);
						neopopZ_asmfastrender((uint8*)tline,(uint8*)fline,(480-SCREEN_WIDTH)<<1);
						stubzodiacblit(3,SCREEN_WIDTH,SCREEN_HEIGHT,5);
						break;
					}
					else if (!neopopZ_smoothmode) {
						tline = ((UInt16*) neopopZ_zodvram);
						neopopZ_asmfastrender((uint8*)tline,(uint8*)fline,(480-SCREEN_WIDTH)<<1);
						stubzodiacblit(3,SCREEN_WIDTH,SCREEN_HEIGHT,1);
						break;
					}
						
				}
				if (device_type&BF_HIRESPLUS){
					tline = ((UInt16*) g_pnobridge -> vram) + ( ((g_pnobridge -> display_height-320)>>1) * display_pitch ) + ((display_width-480)>>1);
					if (neopopZ_smoothmode) fgbZ_filterrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch)<<1),SCREEN_WIDTH*2,SCALE_2XSAI,SCREEN_WIDTH,SCREEN_HEIGHT,480,320);
				}
				break;
		}
	  } else {
	    tline = ((UInt16*) g_pnobridge -> vram) + ( ((g_pnobridge -> display_height-SCREEN_HEIGHT)>>1) * display_pitch ) + ((display_width-SCREEN_WIDTH)>>1);
	    neopopZ_asmfastrender((UInt8*)tline,(UInt8*)fline,(UInt32)((display_pitch-SCREEN_WIDTH)<<1));
	  }
		
//		neopopZ_curRenderingScreenIndex^=1;
//		neopopZ_curRenderingScreenPtr=neopopZ_RenderingScreenPtr[neopopZ_curRenderingScreenIndex];
//		bitmap.data   = (unsigned char *)neopopZ_curRenderingScreenPtr;

	}
	
	//frames_displayed++;	

}


/****************************************************************************/


int neopopZ_ShowMenuGFXConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  int scrmode,smoothing,fskipval,hwsmooth;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(23*sizeof(RectangleType));
  char tmp[16];

	
  scrmode=neopopZ_renderingmode;
  smoothing=neopopZ_smoothmode&_SMOOTH_MASK_;
  hwsmooth=neopopZ_smoothmode&_ZOD_SMOOTHING_;
	
  fskipval=neopopZ_frameskip;

	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {  	  	
		GUI_REDRAW_BG(1);
		
		menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
	  
	  	menu_button(10,80,(char*)"FRAMESKIP",0,&item_rect[1],2);	  
	  	menu_bar(150,80,fskipval,11,&item_rect[2]);
	  	if (fskipval==AUTO_FSKIP) strcpy(tmp,"AUTO");
	  	else if (fskipval==SKEEZIX_FSKIP) strcpy(tmp,"SKZX");
	  	else StrIToA(tmp,fskipval);
	  	menu_messageRaw(260,80,tmp);
	  
	  	menu_button(10,120,(char*)"ZOOM",0,&item_rect[3],2);	  	
	  	menu_button(150,120,(char*)"1",((scrmode==0)||(scrmode==2)?2:1),&item_rect[4],0);
	  	menu_button(200,120,(char*)"2",((scrmode==1)||(scrmode==3)?2:1),&item_rect[5],0);		
	  	menu_button(250,120,(char*)"WIDE",(scrmode==4?2:1),&item_rect[6],0);
	  	
	  	menu_button(10,155,(char*)"FULLSCREEN",0,&item_rect[7],2);
	  	menu_button(150,155,(char*)"OFF",(scrmode&2?1:2),&item_rect[8],0);
	  	menu_button(200,155,(char*)"ON",(scrmode&2?2:1),&item_rect[9],0);	  	  	  
	  
	  	
	  	menu_button(10,190,(char*)"SMOOTHING",0,&item_rect[10],2);
	  	menu_button(150,190,(char*)"OFF",((smoothing==0)&&(!hwsmooth)?2:1),&item_rect[11],0);	  		  	
  		menu_button(200,190,(char*)"AdvMAME2x",(smoothing==1?2:1),&item_rect[12],4);
	  	menu_button(2,218,(char*)"LQ2X",(smoothing==2?2:1),&item_rect[13],4);	  	  	  
	  	menu_button(68,218,(char*)"HQ2X",(smoothing==3?2:1),&item_rect[14],4);
  		menu_button(136,218,(char*)"2xSaI",(smoothing==4?2:1),&item_rect[15],4);
	  	menu_button(202,218,(char*)"Super 2xSaI",(smoothing==5?2:1),&item_rect[16],4);
	  	menu_button(10,246,(char*)"Super Eagle",(smoothing==6?2:1),&item_rect[17],4);
	  	menu_button(130,246,(char*)"Scale 2xSaI",(smoothing==7?2:1),&item_rect[18],4);
  		menu_button(260,246,(char*)"HW",(hwsmooth?2:1),&item_rect[22],4);
	  	
	  	
	  	menu_button(10,280,(char*)"FLIPPING",0,&item_rect[19],2);
		menu_button(150,280,(char*)"OFF",(neopopZ_flipping?1:2),&item_rect[20],0);
	  	menu_button(200,280,(char*)"ON",(neopopZ_flipping?2:1),&item_rect[21],0);
	  	
	  	if (scr) ljp_WinScreenUnlock();
	
  	  }
	  redraw=menu_waitPen(&x,&y,0);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //FSKIP
	  if (TEST_PEN_MENU(x,y,item_rect[2])) {
	  	fskipval=(x-10-item_rect[2].topLeft.x)*11/100;
	  	if (fskipval<0) fskipval=0;
	  	if (fskipval>11) fskipval=11;
	  	redraw=1;
	  }

	  //SCREEN MODES
	  if (TEST_PEN_MENU(x,y,item_rect[4])) {
		  scrmode&=~4;
	  	  scrmode&=~1;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[5])) {
		  scrmode&=~4;
	  	  scrmode|=1;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[6])) {
	  	  scrmode=4;
	  	  redraw=1;
	  }
	  //SCREEN MODES WIN/FULL
	  if (TEST_PEN_MENU(x,y,item_rect[8])) {
	  	  scrmode&=~(2|4);
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) {
	  	  if (scrmode<2) scrmode|=2;
	  	  redraw=1;
	  }
	  //SMOOTHING
	  if (TEST_PEN_MENU(x,y,item_rect[11])) {
	  	  smoothing=0;hwsmooth=0;
  		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[12])) {
  		  smoothing=1;hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[13])) {
	  	  smoothing=2;hwsmooth=0;
 		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[14])) {
  		  smoothing=3;hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[15])) {
	  	  smoothing=4;hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[16])) {
	  	  smoothing=5;
	  	  hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[17])) {
	  	  smoothing=6;
	  	  hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[18])) {
	  	  smoothing=7;
	  	  hwsmooth=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[22])) {
	  	  smoothing=0;
	  	  hwsmooth=_ZOD_SMOOTHING_;
	  	  redraw=1;
	  }
	  //FLIPPING
	  if (TEST_PEN_MENU(x,y,item_rect[20])) {
	  	  neopopZ_flipping=0;
	  	  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[21])) {
	  	  neopopZ_flipping=1;
	  	  redraw=1;
	  }
	  if (mStopEmulation) break;
  }  
  
  neopopZ_frameskip=fskipval;

  neopopZ_renderingmode=scrmode;
  neopopZ_smoothmode=smoothing|hwsmooth;
	
  if (!mStopEmulation) menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}



int neopopZ_ShowMenuSNDConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  int vol;
  int bassboost;
  RectangleType *item_rect;
  char tmp[16];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  
  vol=snd_stream_vol/16;	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG(1)
  		
  	  	  	  	  	  	
	  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(10,80,(char*)"SOUND",0,&item_rect[1],2);
	  menu_button(150,80,(char*)"OFF",(neopopZ_enabledSound==0?2:1),&item_rect[2],0);
	  menu_button(200,80,(char*)"SIM",(neopopZ_enabledSound==1?2:1),&item_rect[3],0);	  
	  menu_button(250,80,(char*)"ON",(neopopZ_enabledSound==2?2:1),&item_rect[4],0);	 
	  
	  menu_button(10,120,(char*)"VOLUME",0,&item_rect[5],2);	  
	  menu_bar(150,120,vol,128,&item_rect[6]);
	  sprintf(tmp,"%d%",vol*100/64);	  	
	  	menu_messageRaw(270,120,tmp);
		if (scr) ljp_WinScreenUnlock();	  
  	  }
	  redraw=menu_waitPen(&x,&y,0);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //SOUND OFF
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
		  neopopZ_enabledSound=0;		  
		  redraw=1;
	  }
	  //SOUND SIM
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
	  	  neopopZ_enabledSound=1;	  	  
	  	  redraw=1;
	  }
	  //SOUND ON
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
		  neopopZ_enabledSound=2;
		  redraw=1;
	  }
	  
	  //VOLUME
	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	vol=(x-item_rect[6].topLeft.x-10)*128/100;
	  	if (vol<0) vol=0;
	  	if (vol>128) vol=128;
	  	redraw=1;
	  }
	  if (mStopEmulation) break;
  }  

	snd_stream_vol=vol*16;
	
	//snd.enabled = neopopZ_enabledSound
	//if (!neopopZ_enabledSound) {
	//	system_init
	//sound_init();
  
  if (!mStopEmulation) menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}


int neopopZ_ShowMenuOTHERConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  char tmp[16];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG  	(1)
  		
  	  	  	  	  	  	
	  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(10,80,(char*)"AUTO A",0,&item_rect[1],4);
	  menu_button(100,80,(char*)"OFF",(neopopZ_autofireA==0?2:1),&item_rect[2],0);
	  menu_button(150,80,(char*)"SLW",(neopopZ_autofireA==1?2:1),&item_rect[3],0);
	  menu_button(200,80,(char*)"MED",(neopopZ_autofireA==2?2:1),&item_rect[4],0);
	  menu_button(250,80,(char*)"FST",(neopopZ_autofireA==3?2:1),&item_rect[5],0);
	  
	  menu_button(10,120,(char*)"AUTO B",0,&item_rect[6],4);
	  menu_button(100,120,(char*)"OFF",(neopopZ_autofireB==0?2:1),&item_rect[7],0);
	  menu_button(150,120,(char*)"SLW",(neopopZ_autofireB==1?2:1),&item_rect[8],0);
	  menu_button(200,120,(char*)"MED",(neopopZ_autofireB==2?2:1),&item_rect[9],0);
	  menu_button(250,120,(char*)"FST",(neopopZ_autofireB==3?2:1),&item_rect[10],0);
	  
	  menu_button(10,160,(char*)"EMU SPEED",0,&item_rect[11],4);
	  menu_bar(130,160,neopopZ_speedpct,100,&item_rect[12]);
	  StrIToA(tmp,neopopZ_speedpct);
	  strcat(tmp,"%");
    	menu_messageRaw(240,160,tmp);
	  
		if (scr) ljp_WinScreenUnlock();	  
  	  }
	  redraw=menu_waitPen(&x,&y,0);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //AUTOFIRE A
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
		  neopopZ_autofireA=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
		  neopopZ_autofireA=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[4])) 
	  {
		  neopopZ_autofireA=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[5])) 
	  {
		  neopopZ_autofireA=3;
		  redraw=1;
	  }
	  //AUTOFIRE B
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
		  neopopZ_autofireB=0;		  
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
		  neopopZ_autofireB=1;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
		  neopopZ_autofireB=2;
		  redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[10])) 
	  {
		  neopopZ_autofireB=3;
		  redraw=1;
	  }
	  //SPEED
	  if (TEST_PEN_MENU(x,y,item_rect[12])) 
	  {
	  	neopopZ_speedpct=(x-item_rect[12].topLeft.x);
	  	if (neopopZ_speedpct>100) neopopZ_speedpct=100;
	  	if (neopopZ_speedpct<10) neopopZ_speedpct=10;
	  	redraw=1;
	  }
	  if (mStopEmulation) break;
  }  
  
  if (!mStopEmulation) menu_waitPen(NULL,NULL,2);
  
  ljz_free(item_rect);
  return result;
}




int neopopZ_ShowMenuSave(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  FILE *savefile;
  	  int slot[4];
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  memset((char*)item_rect,0,15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG  (1)	
  		
  	  	  	  	  	  	
	  
  	  slot[0]=slot[1]=slot[2]=slot[3]=0;
  	  for (int i=0;i<4;i++) 
  	  	if (!state_load(i,1))   	  	
  	  	{
      		char str[255];		
			UInt32 long_time;
			DateTimeType date_time;			
			slot[i]=1;			
			getstate_filename(i,str);		
			fgetdate(str,&long_time);		
			TimSecondsToDateTime(long_time,&date_time);
			
			sprintf(str,"%02d%s%02d-%02d:%02d",
			date_time.day,str_month[date_time.month-1],date_time.year%100,date_time.hour,date_time.minute);
			if (i) 
			{
				menu_messageRaw(150,80+i*30,str);
				menu_button(10,80+i*30,"X",5,&item_rect[6+i],3);
			}
			else 
			{
				menu_messageRaw(150,200,str);
				menu_button(10,200,"X",5,&item_rect[6+i],3);
			}
	  	}	  	  	
  	  	  	  	  	  	
		  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
		  menu_button(10,50,(char*)"SAVE",0,&item_rect[1],2);	  	  	  
		  menu_button(50,110,(char*)"SLOT 1",(slot[1]?2:3),&item_rect[2],1);
		  menu_button(50,140,(char*)"SLOT 2",(slot[2]?2:3),&item_rect[3],1);
		  menu_button(50,170,(char*)"SLOT 3",(slot[3]?2:3),&item_rect[4],1);
		  menu_button(50,200,(char*)"QUICK",(slot[0]?2:3),&item_rect[5],1);
		  if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
  	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	if (menu_confirm("Delete QUICKSLOT?"))
	  	{
		  	char str[255];		
			getstate_filename(0,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	if (menu_confirm("Delete SLOT 1?"))
	  	{
		  	char str[255];		
			getstate_filename(1,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	if (menu_confirm("Delete SLOT 2?"))
	  	{
		  	char str[255];		
			getstate_filename(2,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	if (menu_confirm("Delete SLOT 3?"))
	  	{
		  	char str[255];		
			getstate_filename(3,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //SAVE 0
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  if (!slot[1]||(SHOW_OVERWRITE))
		  {state_save(1);result=1; break;}
		  else redraw=1;
		 
	  }
	  //SAVE 1
	  if (TEST_PEN_MENU(x,y,item_rect[3]))
	  {
		  if (!slot[2]||(SHOW_OVERWRITE))
		  {state_save(2);result=1; break;}
		  else redraw=1;
		 
	  }
	  //SAVE 2
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (!slot[3]||(SHOW_OVERWRITE))
		  {state_save(3);result=1; break;}
		  else redraw=1;
		 
	  }
	  //QUICKSAVE
	  if (TEST_PEN_MENU(x,y,item_rect[5]))
	  {
		  if (!slot[0]||(SHOW_OVERWRITE))
		  {state_save(0);result=1; break;}
		  else redraw=1;
		 
	  }
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}



int neopopZ_ShowMenuLoad(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
   int slot[4];
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
  memset((char*)item_rect,0,15*sizeof(RectangleType));
  
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG  	(1)
		
  	  FILE *savefile;
  	 
  	  slot[0]=slot[1]=slot[2]=slot[3]=0;
      for (int i=0;i<4;i++) 
      	if (!state_load(i,1)) 
      	{
      		char str[255];		
			UInt32 long_time;
			DateTimeType date_time;			
			slot[i]=1;			
			getstate_filename(i,str);		
			fgetdate(str,&long_time);		
			TimSecondsToDateTime(long_time,&date_time);
			
			sprintf(str,"%02d%s%02d-%02d:%02d",
			date_time.day,str_month[date_time.month-1],date_time.year%100,date_time.hour,date_time.minute);
			if (i) 
			{
				menu_messageRaw(150,80+i*30,str);
				menu_button(10,80+i*30,"X",5,&item_rect[6+i],3);
			}
			else 
			{
				menu_messageRaw(150,200,str);
				menu_button(10,200,"X",5,&item_rect[6+i],3);
			}
		}	
      
		  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
		  menu_button(10,50,(char*)"LOAD",0,&item_rect[1],2);	  	  	  
		  menu_button(50,110,(char*)"SLOT 1",(slot[1]?2:3),&item_rect[2],1);
		  menu_button(50,140,(char*)"SLOT 2",(slot[2]?2:3),&item_rect[3],1);
		  menu_button(50,170,(char*)"SLOT 3",(slot[3]?2:3),&item_rect[4],1);
		  menu_button(50,200,(char*)"QUICK",(slot[0]?2:3),&item_rect[5],1);
		  if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;

  	  if (TEST_PEN_MENU(x,y,item_rect[6])) 
	  {
	  	if (menu_confirm("Delete QUICKSLOT?"))
	  	{
		  	char str[255];		
			getstate_filename(0,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[7])) 
	  {
	  	if (menu_confirm("Delete SLOT 1?"))
	  	{
		  	char str[255];		
			getstate_filename(1,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[8])) 
	  {
	  	if (menu_confirm("Delete SLOT 2?"))
	  	{
		  	char str[255];		
			getstate_filename(2,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  if (TEST_PEN_MENU(x,y,item_rect[9])) 
	  {
	  	if (menu_confirm("Delete SLOT 3?"))
	  	{
		  	char str[255];		
			getstate_filename(3,str);
	  		remove(str);
	  	}
	  	redraw=1;
	  }
	  
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //LOAD 0
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  if (slot[1])
		if (!state_load(1,0)) {result=1;break;}
	  }
	  //LOAD 1
	  if (TEST_PEN_MENU(x,y,item_rect[3]))
	  {
		  if (slot[2])
		if (!state_load(2,0)) {result=1;break;}
	  }
	  //LOAD 2
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (slot[3])
		if (!state_load(3,0)) {result=1;break;}	  
	  }
	
	  //QUICKLOAD
	  if (TEST_PEN_MENU(x,y,item_rect[5]))
	  {
	  	if (slot[0])
    	if (!state_load(0,0)) {result=1;break;}
	  }
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}


int neopopZ_ShowMenuConfig(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG  	(1)
  		
  	  	  	  	  	  	
	  menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);	  
	  menu_button(92,100,(char*)"GFX",0,&item_rect[1],2);
	  menu_button(92,150,(char*)"SOUND",0,&item_rect[2],2);
	  menu_button(92,200,(char*)"OTHER",0,&item_rect[3],2);
	  if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}	  
	  //GFX
	  if (TEST_PEN_MENU(x,y,item_rect[1])) 
	  {
	  	neopopZ_ShowMenuGFXConfig(savescr);
		redraw=1;
	  }
	  //SOUND
	  if (TEST_PEN_MENU(x,y,item_rect[2])) 
	  {
     	neopopZ_ShowMenuSNDConfig(savescr);
		redraw=1;
	  }
	  //OTHER
	  if (TEST_PEN_MENU(x,y,item_rect[3])) 
	  {
		  neopopZ_ShowMenuOTHERConfig(savescr);
		  redraw=1;		  
	  }
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}

int neopopZ_ShowSysInfos(uint8 *savescr)
{
  uint8 *scr;
  int x,y;
  int redraw;  
  int result;
  RectangleType *item_rect,my_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));
	    
  redraw=1;         
  for (;;)
  {  
  	  if (redraw)
  	  {
		GUI_REDRAW_BG	(1)			
		
		char str[64],strb[32];
		unsigned long maxP,freeP,usedP;		
		VolumeInfoType vinfo;
		VFSVolumeSize(neopopZ_curVolume,&usedP,&maxP);		
		VFSVolumeGetLabel(neopopZ_curVolume,strb,16);strb[16]=0;
		menu_button(10,70,"Volume",0,&my_rect,4);		
		sprintf(str,"%s[%d]",strb,neopopZ_curVolume);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,63,str);				
		sprintf(str,"Used:%dM/%dM",usedP>>20,maxP>>20);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,81,str);				
		MemHeapFreeBytes(0,&freeP,&maxP);		
		menu_button(10,100,"RAM",6,&my_rect,4);		
		sprintf(str,"RAM %ldK/%ldK",freeP/1024,maxP/1024);
		menu_messageRaw(my_rect.topLeft.x+my_rect.extent.x+5,100,str);		
						  	
	  	menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
		
	  
	  	if (scr) ljp_WinScreenUnlock();
  	  }
	  redraw=menu_waitPen(&x,&y,1);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  
	  //BACK	  
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  
	  if (mStopEmulation) break;
  }  
  ljz_free(item_rect);
  return result;
}



int neopopZ_ShowMenu()
{
  int x,y;
  int redraw;
  int result;
  int slot0;
  uint8 *scr,*savescr,*tmpscr,*tmpdst;
  RectangleType *item_rect;
  item_rect=(RectangleType*)ljz_malloc(15*sizeof(RectangleType));  
  redraw=1;

  
  neopopZ_pause(1);
    
       
  scr=/*(uint8*)(g_pnobridge -> vram);*/ljp_WinScreenLock(winLockCopy);  
  savescr=(uint8*)ljz_malloc(display_width*display_height*2);    
  if (scr&&savescr)
  {
	  tmpscr=scr;
  	for (int yy=0;yy<display_height-(neopopZ_renderingmode<2?BELTBAR_SIZE:0);yy++)
  	{
	  for (x=0;x<display_width;x++)
	  {
  		y=((uint16*)tmpscr)[x];
	  	((uint16*)tmpscr)[x]=((y>>12)<<11)|(((y>>6)&31)<<5)|((y>>1)&15);
	  }
	  tmpscr+=display_pitch*2;
	}
	tmpscr=scr;
	tmpdst=savescr;
	for (int yy=0;yy<display_height;yy++)
  	{
	    memcpy(tmpdst,tmpscr,display_width*2);
	    tmpdst+=display_width*2;
	    tmpscr+=display_pitch*2;
	}
  }
  if (scr) ljp_WinScreenUnlock();
            
              
  for (;;)
  {  
  	if (redraw)
  	{  	
	  	slot0=!state_load(0,1);    
  		
  		
		GUI_REDRAW_BG(1)  		
		
		menu_button(180,20,(char*)"BACK",6,&item_rect[0],2);
		menu_button(10,100,(char*)"SAVE",0,&item_rect[1],2);
		menu_button(180,100,(char*)"QUICK SAVE",(slot0?2:1),&item_rect[2],2);
		menu_button(10,150,(char*)"LOAD",0,&item_rect[3],2);
		menu_button(180,150,(char*)"QUICK LOAD",(slot0?2:3),&item_rect[4],2);
		menu_button(10,200,(char*)"CONFIG",0,&item_rect[5],2);
		menu_button(180,250,(char*)"LAUNCHER",4,&item_rect[6],2);
		
		menu_button(180,200,(char*)"SYS INFOS",7,&item_rect[8],2);
		
		menu_button(10,250,(char*)"RESET NGP",9,&item_rect[7],2);				
		if (scr) ljp_WinScreenUnlock();				
		
  	}
	  redraw=menu_waitPen(&x,&y,1);	  
	  if (redraw==2) {mStopEmulation=1; break;}
	  if (redraw) x=y=-1;
	  //HOME
	  if (TEST_PEN_MENU(x,y,item_rect[6])) {result=1;break;}
	  //BACK
	  if (TEST_PEN_MENU(x,y,item_rect[0])) {result=0;break;}
	  //CONFIG
	  if (TEST_PEN_MENU(x,y,item_rect[5])) {neopopZ_ShowMenuConfig(savescr);redraw=1;}
	  //LOAD
	  if (TEST_PEN_MENU(x,y,item_rect[3])) {if (neopopZ_ShowMenuLoad(savescr)) {result=0;break;} redraw=1;}
	  //SAVE
	  if (TEST_PEN_MENU(x,y,item_rect[1])) {if (neopopZ_ShowMenuSave(savescr)) {result=0;break;} redraw=1;}
	  //QUICK SAVE
	  if (TEST_PEN_MENU(x,y,item_rect[2]))
	  {
		  state_save(0);
		  result=0;
		  break;
	  }
	  //QUICK LOAD
	  if (TEST_PEN_MENU(x,y,item_rect[4]))
	  {
		  if (slot0)
	    if (!state_load(0,0)) {result=0;break;}
	  }
	  //RESET NGP
	  if (TEST_PEN_MENU(x,y,item_rect[7])) {if (menu_confirm("Reset NGP ?")) {reset(); result=0;break;}redraw=1;}
	  
	  //SYS INFOS
	  if (TEST_PEN_MENU(x,y,item_rect[8])) {neopopZ_ShowSysInfos(savescr);redraw=1;}

		if (mStopEmulation) break;
	}
  ljz_free(item_rect);
  
  if (savescr) ljz_free(savescr);
            
  neopopZ_pause(0);
  return result;
}  





//-----------------------------------------------------------------------------
// system_message()
//-----------------------------------------------------------------------------
void __cdecl system_message(char* vaMessage,...)
{
	char message[1001];
    va_list vl;

    va_start(vl, vaMessage);
    vsprintf(message, vaMessage, vl);
	va_end(vl);

/*	if (windowed)
		MessageBox(GetDesktopWindow(), message, PROGRAM_NAME, MB_OK | MB_ICONINFORMATION);*/
	menu_message(message);
//	menu_waitPen(NULL,NULL,1);
}

//=============================================================================

//-----------------------------------------------------------------------------
// windows_load_rom()
//-----------------------------------------------------------------------------
static BOOL windows_load_rom(char* filename)
{
	if (system_load_rom(filename))	//Load the rom
	{
		//Add to MRU

		reset();
		return TRUE;
	}

	return FALSE;
}



//-----------------------------------------------------------------------------
// system_VBL()
//-----------------------------------------------------------------------------

void system_VBL(void)
{
	neopopZ_framecount++;
	//Throttle speed
	int haswaited;
	neopopZ_frameBeginTime+=neopopZ_frameTime*100/neopopZ_speedpct;	//60Hz
			
	//Frameskip
	if ((neopopZ_frameskip==AUTO_FSKIP)&&(!neopopZ_fastforward))
	{
		
		//do avoid issue when Tick is looping
		if ((_s32)(neopopZ_frameBeginTime-GetTickCount())>=(_s32)(20*neopopZ_frameTime))	
			neopopZ_frameBeginTime=GetTickCount()+neopopZ_frameTime*100/neopopZ_speedpct;				
		while (GetTickCount()<=neopopZ_frameBeginTime)  haswaited=1;		
		/*int i=neopopZ_frameBeginTime-GetTickCount()+1;
        if (i>0) {haswaited=1; SysTaskDelay(i);}*/

	
		if (haswaited) 
		{
			neopopZ_frameskipped=0;
			neopopZ_drawframe=1;
		}
		else 
		{
			neopopZ_frameskipped++;
			if (neopopZ_frameskipped>MAX_FRAME_SKIPPED) 
			{
				neopopZ_frameskipped=0;
				neopopZ_drawframe=1;
			}
			else neopopZ_drawframe=0;
		}			
	}
	else
	{
		if (neopopZ_frameskip==SKEEZIX_FSKIP) neopopZ_frameskipped=SKEEZIX_FSKIP;
		neopopZ_frameskipped++;
		if (neopopZ_fastforward)
		{
			if (neopopZ_frameskipped>NEOPOPZ_TURBOFSKIP)
			{
				neopopZ_frameskipped=0;
				neopopZ_drawframe=1;
			}
			else neopopZ_drawframe=0;
		}
		else
		{
			if (neopopZ_frameskipped>neopopZ_frameskip)
			{
				neopopZ_frameskipped=0;
				neopopZ_drawframe=1;
			}
			else neopopZ_drawframe=0;
		}

		if ((!neopopZ_fastforward)&&(neopopZ_drawframe||(neopopZ_enabledSound==2)))
		{
			//do avoid issue when Tick is looping
			if ((_s32)(neopopZ_frameBeginTime-GetTickCount())>=(_s32)(20*neopopZ_frameTime))	
				neopopZ_frameBeginTime=GetTickCount()+neopopZ_frameTime*100/neopopZ_speedpct;				
			while (GetTickCount()<=neopopZ_frameBeginTime) ;
			/*int i=neopopZ_frameBeginTime-GetTickCount()+1;
	        if (i>0) {SysTaskDelay(i);}*/
		}
	}
	
	if (neopopZ_drawframe)	// Update Graphics
		neopopZ_updateScreen();

	system_input_update();	
}
