//---------------------------------------------------------------------------
// NEOPOP : Emulator as in Dreamland
//
// Copyright (c) 2001-2002 by neopop_uk
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version. See also the license.txt file for
//	additional informations.
//---------------------------------------------------------------------------

/*
//---------------------------------------------------------------------------
//=========================================================================

	system_sound.c

//=========================================================================
//---------------------------------------------------------------------------

  History of changes:
  ===================

20 JUL 2002 - neopop_uk
=======================================
- Cleaned and tidied up for the source release

26 JUL 2002 - neopop_uk
=======================================
- Offset the initial write pointer to prevent corruption.
	Seems to have improved things quite a bit.

31 JUL 2002 - neopop_uk
=======================================
- Converted sound buffer to use unsigned data.

01 AUG 2002 - neopop_uk
=======================================
- Changed the sound buffer length to be an integral number of frames.
- Added a #define for the mix-ahead value (write pointer offset)

03 AUG 2002 - neopop_uk
=======================================
- Changed the DAC code so that it uses a seperate buffer, this way
the frequency can be controlled precisely. 8 khz sounds right.
- DAC buffer is 8 bit. Chip buffer is still 16 bit.
- Converted DAC to mono, i've never seen it used in stereo.

08 AUG 2002 - neopop_uk
=======================================
- Re-written buffering using dynamic block length
- Replaced sound start and stop with 'system_sound_clear'.
- system_sound_init() now calls system_sound_clear()

08 AUG 2002 - neopop_uk
=======================================
- Fixed a crash bug by making system_sound_shutdown set ptr's to NULL
- Fixed the DAC buffer size

09 AUG 2002 - neopop_uk
=======================================
- Reduced latency.
- Optimised the UNDEFINED state to ensure DAC stays in sync.
- The chip sound buffer is always stereo format.

16 AUG 2002 - neopop_uk
=======================================
- The audio stream generation is always mono, but I've had
better results with a stereo buffer, so the data is doubled up.

//---------------------------------------------------------------------------
*/

#include "neopop.h"


#include "system_main.h"
#include "system_sound.h"
#include "system_config.h"

//=============================================================================


#include "SoundMgr.h"
SndStreamRef snd_stream;
typedef struct {_u32 __r9;_u32 __r10;} reg_sys_t;
static reg_sys_t reg_sys;
_u32 neopopZ_sndbuffsize;
extern _u32 neopopZ_sndfreq,neopopZ_enabledSound,snd_stream_vol;
_s32 snd_cb_running;
extern _s32 neopopZ_frameBeginTime,neopopZ_frameTime,neopopZ_speedpct;

static Err neopopZ_sndCallBack (void *userData,SndStreamRef stream, void *bufferZ,UInt32 frameCountZ)
{
	__asm 
	{
		stmfd  r13!,{r9,r10}
		ldr	   r9,[r0]
		ldr	   r10,[r0,#4]
	}		
	if (snd_cb_running==0)
	{
		snd_cb_running=1;
		sound_update((_s16*)(bufferZ),frameCountZ);
		snd_cb_running=0;
	}
	
	__asm 
	{
		ldmfd  r13!,{r9,r10}
	}
	return errNone;
}


//=============================================================================

/*void system_sound_update(void)
{
}*/

//=============================================================================

BOOL system_sound_init(void)
{
	int bufflen;
	switch (neopopZ_sndfreq/1000)
	{
		case 11:neopopZ_sndbuffsize=512;break;
		case 22:neopopZ_sndbuffsize=512;break;
		case 44:neopopZ_sndbuffsize=1024;break;
	}
	
    __asm 
    {
  		ldr  r0, = reg_sys
	  	add	 r0,r0,r10
  		str	 r9,[r0]
	  	str	 r10,[r0,#4]
    } 
    snd_cb_running=0;    
	
	return TRUE;
}

//=============================================================================

void system_sound_chipreset(void)
{
	//Initialises sound chips, matching frequncies
	sound_init(neopopZ_sndfreq);
}
	
void system_sound_silence(void)
{

}

//=============================================================================

void system_sound_shutdown(void)
{

}

//=============================================================================


void neopopZ_pause(int pause)
{	
   if (pause)
	{
		if (snd_stream) stub_SndStreamGetVolume(snd_stream,&snd_stream_vol);  
		if (neopopZ_enabledSound==2) 
		{
			stub_SndStreamStop(snd_stream);
			stub_SndStreamDelete(snd_stream);  
		}
	}
	else
	{
		neopopZ_frameBeginTime=GetTickCount()+neopopZ_frameTime*100/neopopZ_speedpct;
		if (neopopZ_enabledSound==2) 
		{
			stub_SndStreamCreate(&snd_stream,sndOutput,neopopZ_sndfreq,sndInt16Little,sndMono,neopopZ_sndCallBack,&reg_sys,neopopZ_sndbuffsize,1);
			stub_SndStreamStart(snd_stream);		
			stub_SndStreamSetVolume(snd_stream,snd_stream_vol);
		}
	}
}
