//---------------------------------------------------------------------------------------
// make sound buffer
//---------------------------------------------------------------------------------------

#if defined(SOUND)

#include "stdinc.h"
#include "string.h"
#include "syscall.h"
#include "pce.h"
#include "sound.h"
#include "pg.h"
#include "main.h"

void   pgaSetChannelCallback(int channel, void *callback);

static int mseq(DWORD *rand_val);
int    WriteBuffer2(int ch, DWORD dwSize);
void   WriteSoundData(int ch, DWORD dwSize);

// internal 
static int CycleLimit;            // (1byte)ɕKvȍŏClockl (7160000/44100=162.3582...)
static int sound_enable=1;        // 0:Sound Off / 1:Sound On

#define    SAMPLE_RATE    ((DWORD)44100)// sampling rate
#define    SAMPLE10       (SAMPLE_RATE*10)

#define    SND_L            1
#define    SND_R            0

//------------------------------------------------------------------------------
// Buffer Size
//------------------------------------------------------------------------------
#define    SND_FRMSIZE    736     // 1 frame sound data size (=44100/59.94)
#define    SND_BNKSIZE    512     // buffer read size (see pg.c)

#define    SND_UNDERLIMIT   (2*SND_FRMSIZE)          // underflow limit
#define    SND_OVERLIMIT    (10*SND_FRMSIZE)         // overflow  limit
#define    SND_RNGSIZE      (40*SND_BNKSIZE)         // sound ring buffer 

//------------------------------------------------------------------------------
// last cycle timing
//------------------------------------------------------------------------------
static DWORD dwOldPos[6]={0,0,0,0,0,0};
static DWORD CycleOld[6]={0,0,0,0,0,0};

//------------------------------------------------------------------------------
// 16bit Sound Ring Buffer
//------------------------------------------------------------------------------
static int   snd_wr=0;                  // Sound Write Pointer
static int   snd_rd=0;                  // Sound Read  Pointer
static short sndbuffer[SND_RNGSIZE][2]; // Sound Ring Buffer
static int   over_wr=0,over_rd=0;       // (Debug) overrun counter
static int   unlk_wr=0,unlk_rd=0;       // (Debug) unlock error

//------------------------------------------------------------------------------
// debug 
//------------------------------------------------------------------------------
static int   snd_sound_skip=0;          // sound write skip (overflow limit check)
static int   snd_vsync_skip=0;          // underflow vsync skip counter
static int   snd_frame_skip=0;          // underflow frame skip counter

//------------------------------------------------------------------------------
// Sound Mix Buffer
// 1/60bɍRing Bufferֈړ
//------------------------------------------------------------------------------
static int  mix_wave[SND_FRMSIZE*4][2]; // 1/60ɕKvbuffer x 4 
static int  mix_ch[6]={0,0,0,0,0,0};    // ςindexl(ch)

// internal function
static int    bufLen(void);
static short *bufSetLock(int size);
static void   bufSetUnlock(void* p,int size);
static short *bufGetLock(int size);
static void   bufGetUnlock(void* p,int size);

//------------------------------------------------------------------------------
// rest table
//------------------------------------------------------------------------------
void reset_sound(void)
{
    _memset(dwOldPos,0,sizeof(dwOldPos));
    _memset(CycleOld,0,sizeof(CycleOld));
    _memset(mix_wave,0,sizeof(mix_wave));
    _memset(mix_ch  ,0,sizeof(mix_ch)  );
}

//------------------------------------------------------------------------------
// enable or disable
//------------------------------------------------------------------------------
void enable_sound(int mode)
{
    if(mode) {
        if(sound_enable==0) {
            reset_sound();
        }
        sound_enable=1;
    }
    else {
        sound_enable=0;
    }
}


//------------------------------------------------------------------------------
// Init Parameter
//------------------------------------------------------------------------------
void initSound(void)
{
    snd_wr=snd_rd=0;
    _memset(sndbuffer,0,sizeof(sndbuffer));
    _memset(CycleOld ,0,sizeof(CycleOld) );
    _memset(dwOldPos ,0,sizeof(dwOldPos) );
    _memset(mix_wave ,0,sizeof(mix_wave) );
    _memset(mix_ch   ,0,sizeof(mix_ch)   );
}

//------------------------------------------------------------------------------
// Cleanup Parameter
//------------------------------------------------------------------------------
void closeSound(void)
{
    
}

//------------------------------------------------------------------------------
// buffer̒Qbc
//------------------------------------------------------------------------------
static int bufLen(void)
{
    register int rd=snd_rd;
    register int wr=snd_wr;
    if(wr==rd) return 0;
    if(wr >rd) return wr-rd;;
    return SND_RNGSIZE - rd + wr;
}

//------------------------------------------------------------------------------
// ɏމf[^̈Qbc֐
// ۂɃQbcAhX͖ߒlgĂق
//------------------------------------------------------------------------------
static short *bufSetLock(int size)
{
    int p = (snd_wr+size)%SND_RNGSIZE;

    if(p!=snd_rd) {
        return sndbuffer[snd_wr];
    }
    over_wr++;
    
    return 0;
}

//------------------------------------------------------------------------------
// [Write] bN
//------------------------------------------------------------------------------
static void bufSetUnlock(void*ptr,int size)
{
    if( ptr==(void*)&sndbuffer[snd_wr] ) {
        snd_wr=(snd_wr+size)%SND_RNGSIZE;
    } else {
        unlk_wr++;
    }
}

//------------------------------------------------------------------------------
// Sound Ring Buffer特f[^E
//------------------------------------------------------------------------------
static short *bufGetLock(int size)
{
    if(bufLen()>=size) { // Kvȏ̃f[^邩H
        return sndbuffer[snd_rd];
    }
    over_rd++;
    return 0;
}

//------------------------------------------------------------------------------
// [Read] 
//------------------------------------------------------------------------------
static void bufGetUnlock(void* ptr,int size)
{
    if(ptr==(void*)&sndbuffer[snd_rd]) {
        snd_rd=(snd_rd+size)%SND_RNGSIZE;
    } else {
        unlk_rd++;
    }
}


extern int skip_next_frame;
extern int skip_next_vsync;

int prev_skip_frame=0;

//------------------------------------------------------------------------------
//ÿ艻z
// TEh쐬ԂɍȂꍇ...
// (1)VSYNCOFFɂĂ݂B
// (2)FRAMESKIPONɂĂ݂B
// ȏ̂Q_Ŋ撣Ă݂B
//------------------------------------------------------------------------------
int SoundStabilizer(void)
{
    if(bufLen()<=SND_UNDERLIMIT) {
        if(skip_next_vsync==0) {
            skip_next_vsync++;
            snd_vsync_skip++;
        } else {
            skip_next_frame++;
            snd_frame_skip++;
        }
    } else {
        skip_next_vsync=0;
        skip_next_frame=0;
    }
    
    // Oframe skipȂframe skipȂB
    // skipAƉʕ\ꂸQ[ɂȂȂ
    if(prev_skip_frame) {
        skip_next_frame=0;
    }
    
    prev_skip_frame = skip_next_frame;
    
    return ((skip_next_frame+skip_next_vsync)>0);
}


//------------------------------------------------------------------------------
// 1/60bԊuŎsė~֐
// 6ch̉
//------------------------------------------------------------------------------
void SoundProcessing(void)
{
    if(sound_enable) {
        // calc sound data
        WriteSoundData(0,SND_FRMSIZE); 
        WriteSoundData(1,SND_FRMSIZE);
        WriteSoundData(2,SND_FRMSIZE);
        WriteSoundData(3,SND_FRMSIZE);
        WriteSoundData(4,SND_FRMSIZE);
        WriteSoundData(5,SND_FRMSIZE);
        
#ifdef MP3
        mp3decode();
#endif
        
        // convert int to short
        {
            register int i;
            register int wave;
            register int swr = snd_wr;
            
#ifdef MP3
            register short *mp3ptr = mp3callback(0,SND_FRMSIZE);
#endif//MP3
            for(i=0;i<SND_FRMSIZE;i++) {
                wave = mix_wave[i][SND_R];
#ifdef MP3
                wave += *mp3ptr;
#endif

                if(wave > 0x7FFF)      wave =  0x07FFF;
                else if(wave< -0x8000) wave = -0x08000;
                
                sndbuffer[swr][SND_R] = (short)wave;
                mix_wave[i][SND_R]=0;

                wave = mix_wave[i][SND_L];
#ifdef MP3
                wave += *mp3ptr;
                mp3ptr+=2;
#endif
                
                if(wave > 0x7FFF)      wave =  0x07FFF;
                else if(wave< -0x8000) wave = -0x08000;

                sndbuffer[swr][SND_L] = (short)wave;
                mix_wave[i][SND_L]=0;
                swr++;

                if(swr>=SND_RNGSIZE){
                    swr=0;
                }
            }

            // buffer overflow limit Ȃǉ
            if(bufLen()<SND_OVERLIMIT) {
                snd_wr = swr;
            }else {
                snd_sound_skip++;
            }
            
            mix_ch[0]=mix_ch[1]=mix_ch[2]=
            mix_ch[3]=mix_ch[4]=mix_ch[5]=0;
            
            if(pgPSP()){
                sceKernelWakeupThread(pga_threadhandle[0]);
            }
        }
    }
}


//------------------------------------------------------------------------------
// Emulation corepsgɕωƂf[^
// ƂĂ΂֐
//------------------------------------------------------------------------------
void write_psg(int ch)
{
    if(sound_enable) {
        register DWORD dwNewPos;
        register int Cycle = (DWORD)M.User - CycleOld[ch];
        
    	// I[o[t[^C~OŕςɂȂ邩Ȃ
        if(Cycle<0) {
            CycleOld[ch] = (DWORD)M.User;
        } else {
            // TEhf[^WriteԊu臒l𒴂Ă邩H
            if(Cycle>=CycleLimit) {
                dwNewPos = Cycle/CycleLimit;
                WriteBuffer2(ch, (dwNewPos-dwOldPos[ch]));
                dwOldPos[ch] = dwNewPos;
            }
        }
    }
}

//------------------------------------------------------------------------------
//
//
//------------------------------------------------------------------------------
static int mseq(DWORD *rand_val)
{
	if (*rand_val & 0x00080000)	{
		*rand_val = ((*rand_val ^ 0x0004) << 1) + 1;
		return 1;
	} else {
		*rand_val <<= 1;
		return 0;
	}
}

//------------------------------------------------------------------------------
// dwSize1ch̃TCYw肷邱
// CH̑ǗWriteBufferŎ߂
//------------------------------------------------------------------------------
void WriteSoundData(int ch, DWORD dwNewPos)
{
    if(dwOldPos[ch] < dwNewPos){
        WriteBuffer2(ch, (dwNewPos-dwOldPos[ch]));
    }
    
    CycleOld[ch] = (DWORD)M.User;
    dwOldPos[ch] = 0;
}


//
// WX^g߂ƒxȂꍇ݂Ȃ̂Œ
//
int WriteBuffer2(int ch, DWORD dwSize)
{
	static DWORD n[6] = {0,0,0,0,0,0};
	static DWORD k[6] = {0,0,0,0,0,0};
	static DWORD r[6];
	static DWORD rand_val[6] = {0,0,0,0,0x51F631E4,0x51F631E4};
	/* a*2^(b*x) */
	/* a = 400, b = log2(16384/a), x = 0.0 ~ 1.0 */
	static int	 vol_tbl[32] = {
		 100, 451, 508,  573,  646,  728,  821,  925,
		1043,1175,1325, 1493, 1683, 1898, 2139, 2411,
		2718,3064,3454, 3893, 4388, 4947, 5576, 6285,
		7085,7986,9002,10148,11439,12894,14535,16384,
	};
    register int lvol, rvol;
    
    // ݃TCYI[o[v͐؂l߂
    if((dwSize+mix_ch[ch])>(SND_FRMSIZE)) {
        dwSize = 0; // (SND_FRMSIZE)-mix_ch[ch];
    }

    if(dwSize==0) {
        return mix_ch[ch];
    }
    
	if (!(io.PSG[ch][4]&0x80)) {
		n[ch] = k[ch] = 0;
        mix_ch[ch]+=dwSize;
        return mix_ch[ch];
	}

    /* make volume data from table & register*/
    {
        int psgv = io.psg_volume;
        int psg4 = io.PSG[ch][4] & 0x1f;
        int psg5 = io.PSG[ch][5];
        
        lvol = ((psgv>>3)&0x1E) + psg4 + ((psg5>>3)&0x1E) - 60;
        if(lvol<0) lvol = vol_tbl[0];
        else       lvol = vol_tbl[lvol];
        
        rvol = ((psgv<<1)&0x1E) + psg4 + ((psg5<<1)&0x1E) - 60;
        if (rvol<0) rvol = vol_tbl[0];
        else        rvol = vol_tbl[rvol];
    }

    // Direct D/A output
    if (io.PSG[ch][4]&0x40) {
        register int wav0 = (((int)io.wave[ch][0])-16)*702;
        register int mixch=mix_ch[ch];
        
        lvol = (int)wav0*lvol>>14;
        rvol = (int)wav0*rvol>>14;
        
        for(;dwSize;dwSize--) {
            mix_wave[mixch][SND_L] += lvol;
            mix_wave[mixch][SND_R] += rvol;
            mixch++;
        }
        
        mix_ch[ch]=mixch;
        
	} else {
        if (ch >= 4 && (io.PSG[ch][7]&0x80)) {
            register DWORD ra = r[ch];
            register DWORD ka = k[ch];
            int ra7020l=(7020*lvol)>>14;
            int ra7020r=(7020*rvol)>>14;
            register int sndL=(ra?ra7020l:-ra7020l);
            register int sndR=(ra?ra7020r:-ra7020r);
            register int mixch=mix_ch[ch];
            register DWORD Np = (io.PSG[ch][7]&0x1F);
            register DWORD t;
            
            Np = 3000 + (Np<<9);
            
            for(;dwSize;dwSize--) {
                ka += Np;
                t = ka / 44100;
                if (t >= 1) {
                    ra = mseq(&rand_val[ch]);
                    ka -= 44100*t;
                    sndL=(ra?+ra7020l:-ra7020l);
                    sndR=(ra?+ra7020r:-ra7020r);
                }
                mix_wave[mixch][SND_L]+= sndL;
                mix_wave[mixch][SND_R]+= sndR;
                mixch++;
            }
            
            r[ch]=ra;
            k[ch]=ka;
            mix_ch[ch]=mixch;
            return mix_ch[ch];
        }
        else {
            DWORD Tp = io.PSG[ch][2]+((DWORD)io.PSG[ch][3]<<8);
            int i;
            DWORD t;

            if (Tp<2) {
                mix_ch[ch]+=dwSize;
                return mix_ch[ch];
            }

            Tp-=1;
            
            {
#if 1
                static short wave[32];
                for(i=0;i<32;i++) {
                    wave[i] = ((short)io.wave[ch][i]-16)*702;
                }
#else
                static WORD wave[32];
                for(i=0;i<32;i++) {
                    wave[i] = ((short)io.wave[ch][i]-16)*702;
                }
#endif
                
                {
                    register DWORD na = n[ch];
                    register DWORD ka = k[ch];
                    register DWORD NTp = (32*1118608)/Tp;
                    register int wavena;
                    register int mixch = mix_ch[ch];
                    
                    for(;dwSize;dwSize--) {
                        wavena = wave[na];
                        mix_wave[mixch][SND_L] += ((int)(short)wavena*lvol>>14);
                        mix_wave[mixch][SND_R] += ((int)(short)wavena*rvol>>14);
                        mixch++;
                        ka += NTp;
                        t = ka/SAMPLE10;
                        na = (na+t)&31;
                        ka -= SAMPLE10*t;
                    }
                    
                    n[ch]=na;
                    k[ch]=ka;
                    mix_ch[ch]=mixch;
                }
            }
        }
    }
    return mix_ch[ch];
}

//------------------------------------------------------------------------------
// Sound Player Thread main function
//------------------------------------------------------------------------------
void* wavout_snd0_callback(short *buf, unsigned long reqn)
{
    static short *src=0;
    int buflen;

    if(sound_enable) {
        if(src) {
            bufGetUnlock(src,reqn);
            src=0;
        }
    
        for(buflen=bufLen();
            buflen<(reqn*2); // (2005.07.01) ̌H
            buflen=bufLen() ) {
            sceKernelSleepThread(); // if no sound buffer, sleep...
        }
        src = bufGetLock(reqn);
        return src;
    }

    return 0;
}

//------------------------------------------------------------------------------
//
// Sound Initialize
// 
//------------------------------------------------------------------------------
int wavoutInit(void)
{
    // Sound data 1bytêɕKvCycleԊu
    CycleLimit = BaseClock / SAMPLE_RATE;

    // bufSetLock(SND_FRMSIZE * SND_CHMAX);
    // setup thread function
    pgaSetChannelCallback(0,wavout_snd0_callback);
    return 0;
}

//------------------------------------------------------------------------------
//
// Sound Close
// 
//------------------------------------------------------------------------------
int wavoutClose(void)
{
    return 1;
}


extern int ch5p;
extern int ch5wav[32];

//------------------------------------------------------------------------------
// fobOp̕\Ȃ
//------------------------------------------------------------------------------
void dbgSound(void)
{
    /*if(eConf.debug==2)*/ if (0) {
        int i,x;
        
        for(i=0;i<8;i++){
            x=25*i;
            mh_print_hex4(x,180,io.PSG[0][i],-1); 
            mh_print_hex4(x,190,io.PSG[1][i],-1); 
            mh_print_hex4(x,200,io.PSG[2][i],-1); 
            mh_print_hex4(x,210,io.PSG[3][i],-1); 
            mh_print_hex4(x,220,io.PSG[4][i],-1); 
            mh_print_hex4(x,230,io.PSG[5][i],-1); 
            //mh_print_hex4(x,240,io.PSG[6][i],-1); 
            //mh_print_hex4(x,250,io.PSG[7][i],-1); 
        }
        x+=25;
        mh_print_hex4(x,240,io.psg_lfo_freq,-1); 
        mh_print_hex4(x,250,io.psg_lfo_ctrl,-1); 
        
        for(i=0;i<32;i++) {
            mh_print_hex2(i*12+40, 0,io.wave[0][i],-1);
            mh_print_hex2(i*12+40,10,io.wave[1][i],-1);
            mh_print_hex2(i*12+40,20,io.wave[2][i],-1);
            mh_print_hex2(i*12+40,30,io.wave[3][i],-1);
            mh_print_hex2(i*12+40,40,io.wave[4][i],-1);
            mh_print_hex2(i*12+40,50,io.wave[5][i],-1);
            mh_print_hex2(i*12+40,70,ch5wav[i],(ch5p==i)?RGB_BLUE:-1);
        }
        
        mh_print_hex4(0,140,io.wavofs[4],-1);
        mh_print_hex4(0,150,io.wavofs[5],-1);
    }
    else {
        mh_print_hex4(0,20,snd_vsync_skip,-1);
        mh_print_hex4(0,30,snd_frame_skip,-1);
        
  
    }
}


#endif // defined(SOUND)

